from __future__ import annotations

import json
import logging
import urllib.request


class ProxyHandler:
    current_proxy_url: str | None = None
    log = logging.getLogger("maufbapi.proxy")

    def __init__(self, api_url: str | None) -> None:
        self.api_url = api_url

    def get_proxy_url_from_api(self) -> str | None:
        assert self.api_url is not None

        request = urllib.request.Request(self.api_url, method="GET")
        self.log.debug("Requesting proxy from: %s", self.api_url)

        try:
            with urllib.request.urlopen(request) as f:
                response = json.loads(f.read().decode())
        except Exception:
            self.log.exception("Failed to retrieve proxy from API")
        else:
            return response["proxy_url"]

        return None

    def update_proxy_url(self) -> bool:
        old_proxy = self.current_proxy_url
        new_proxy = None

        if self.api_url is not None:
            new_proxy = self.get_proxy_url_from_api()
        else:
            new_proxy = urllib.request.getproxies().get("http")

        if old_proxy != new_proxy:
            self.log.debug("Set new proxy URL: %s", new_proxy)
            self.current_proxy_url = new_proxy
            return True

        self.log.debug("Got same proxy URL: %s", new_proxy)
        return False

    def get_proxy_url(self) -> str | None:
        if not self.current_proxy_url:
            self.update_proxy_url()

        return self.current_proxy_url



# mautrix-instagram - A Matrix-Instagram puppeting bridge.
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import List, Optional, Union
import logging

from attr import dataclass
import attr

from mautrix.types import (
    JSON,
    ExtensibleEnum,
    Obj,
    SerializableAttrs,
    SerializableEnum,
    SerializerError,
)
from mautrix.types.util.serializable_attrs import _dict_to_attrs

from .account import BaseResponseUser, UserIdentifier

log = logging.getLogger("mauigpapi.types")


class ThreadItemType(ExtensibleEnum):
    DELETION = "deletion"
    MEDIA = "media"
    TEXT = "text"
    LIKE = "like"
    HASHTAG = "hashtag"
    PROFILE = "profile"
    MEDIA_SHARE = "media_share"
    CONFIGURE_PHOTO = "configure_photo"
    CONFIGURE_VIDEO = "configure_video"
    SHARE_VOICE = "share_voice"
    LOCATION = "location"
    ACTION_LOG = "action_log"
    TITLE = "title"
    USER_REACTION = "user_reaction"
    HISTORY_EDIT = "history_edit"
    REACTION_LOG = "reaction_log"
    REEL_SHARE = "reel_share"
    DEPRECATED_CHANNEL = "deprecated_channel"
    LINK = "link"
    RAVEN_MEDIA = "raven_media"
    LIVE_VIDEO_SHARE = "live_video_share"
    TEST = "test"
    STORY_SHARE = "story_share"
    REEL_REACT = "reel_react"
    LIVE_INVITE_GUEST = "live_invite_guest"
    LIVE_VIEWER_INVITE = "live_viewer_invite"
    TYPE_MAX = "type_max"
    PLACEHOLDER = "placeholder"
    PRODUCT = "product"
    PRODUCT_SHARE = "product_share"
    VIDEO_CALL_EVENT = "video_call_event"
    POLL_VOTE = "poll_vote"
    FELIX_SHARE = "felix_share"
    ANIMATED_MEDIA = "animated_media"
    CTA_LINK = "cta_link"
    VOICE_MEDIA = "voice_media"
    STATIC_STICKER = "static_sticker"
    AR_EFFECT = "ar_effect"
    SELFIE_STICKER = "selfie_sticker"
    REACTION = "reaction"
    CLIP = "clip"
    GUIDE_SHARE = "guide_share"
    XMA_MEDIA_SHARE = "xma_media_share"
    XMA_REEL_SHARE = "xma_reel_share"
    XMA_STORY_SHARE = "xma_story_share"
    XMA_REEL_MENTION = "xma_reel_mention"


@dataclass(kw_only=True)
class ThreadItemActionLog(SerializableAttrs):
    description: str
    # TODO bold, text_attributes


class ViewMode(SerializableEnum):
    ONCE = "once"
    REPLAYABLE = "replayable"
    PERMANENT = "permanent"


@dataclass(kw_only=True)
class CreativeConfig(SerializableAttrs):
    capture_type: str
    camera_facing: Optional[str] = None
    should_render_try_it_on: Optional[bool] = None


@dataclass(kw_only=True)
class CreateModeAttribution(SerializableAttrs):
    type: str
    name: str


@dataclass(kw_only=True)
class ImageVersion(SerializableAttrs):
    width: int
    height: int
    url: str
    estimated_scan_sizes: Optional[List[int]] = None


@dataclass(kw_only=True)
class ImageVersions(SerializableAttrs):
    candidates: List[ImageVersion]


@dataclass(kw_only=True)
class VideoVersion(SerializableAttrs):
    type: int
    width: int
    height: int
    url: str
    id: Optional[str] = None


class MediaType(SerializableEnum):
    IMAGE = 1
    VIDEO = 2
    AD_MAP = 6
    LIVE = 7
    CAROUSEL = 8
    LIVE_REPLAY = 9
    COLLECTION = 10
    AUDIO = 11
    SHOWREEL_NATIVE = 12

    @property
    def human_name(self) -> str:
        return self.name.lower().replace("_", " ")


@dataclass(kw_only=True)
class ExpiredMediaItem(SerializableAttrs):
    media_type: Optional[MediaType] = None
    user: Optional[BaseResponseUser] = None


@dataclass(kw_only=True)
class RegularMediaItem(SerializableAttrs):
    id: str
    image_versions2: Optional[ImageVersions] = None
    video_versions: Optional[List[VideoVersion]] = None
    original_width: Optional[int] = None
    original_height: Optional[int] = None
    media_type: MediaType
    media_id: Optional[int] = None
    organic_tracking_token: Optional[str] = None
    creative_config: Optional[CreativeConfig] = None
    create_mode_attribution: Optional[CreateModeAttribution] = None
    is_commercial: Optional[bool] = None
    commerciality_status: Optional[str] = None  # TODO enum? commercial

    @property
    def best_image(self) -> Optional[ImageVersion]:
        if not self.image_versions2:
            return None
        best: Optional[ImageVersion] = None
        for version in self.image_versions2.candidates:
            if version.width == self.original_width and version.height == self.original_height:
                return version
            elif not best or (version.width * version.height > best.width * best.height):
                best = version
        return best

    @property
    def best_video(self) -> Optional[VideoVersion]:
        if not self.video_versions:
            return None
        best: Optional[VideoVersion] = None
        for version in self.video_versions:
            if version.width == self.original_width and version.height == self.original_height:
                return version
            elif not best or (version.width * version.height > best.width * best.height):
                best = version
        return best


@dataclass(kw_only=True)
class Caption(SerializableAttrs):
    pk: int
    user_id: int
    text: str
    # TODO enum? 1
    type: int
    created_at: int
    created_at_utc: int
    # TODO enum? comment
    content_type: str
    # TODO enum? Active
    status: str
    # TODO enum-ish thing?
    bit_flags: int
    user: BaseResponseUser
    did_report_as_spam: bool
    share_enabled: bool
    media_id: int

    # Might not be in all captions
    is_covered: Optional[bool] = None
    private_reply_status: Optional[int] = None


@dataclass
class Location(SerializableAttrs):
    pk: int
    short_name: str
    facebook_places_id: int
    # TODO enum?
    external_source: str  # facebook_places
    name: str
    address: str
    city: str
    lng: Optional[float] = None
    lat: Optional[float] = None
    is_eligible_for_guides: bool = False


@dataclass(kw_only=True)
class CarouselMediaItem(RegularMediaItem, SerializableAttrs):
    carousel_parent_id: str
    pk: int


@dataclass
class UserTag(SerializableAttrs):
    user: BaseResponseUser
    position: List[float]
    # start_time_in_video_in_sec
    # duration_in_video_in_sec


@dataclass
class UserTags(SerializableAttrs):
    in_: List[UserTag] = attr.ib(metadata={"json": "in"}, factory=lambda: [])


@dataclass(kw_only=True)
class MediaShareItem(RegularMediaItem, SerializableAttrs):
    taken_at: int
    pk: int
    device_timestamp: int
    code: str
    client_cache_key: str
    filter_type: int
    user: BaseResponseUser
    # Not present in reel shares
    can_viewer_reshare: Optional[bool] = None
    caption_is_edited: bool = False
    comment_likes_enabled: bool = False
    comment_threading_enabled: bool = False
    has_more_comments: bool = False
    max_num_visible_preview_comments: int = 0
    # preview_comments: List[TODO]
    can_view_more_preview_comments: bool = False
    comment_count: int = 0
    like_count: int = 0
    has_liked: bool = False
    photo_of_you: bool = False
    usertags: Optional[UserTags] = None
    caption: Optional[Caption] = None
    can_viewer_save: bool = True
    location: Optional[Location] = None
    carousel_media_count: Optional[int] = None
    carousel_media: Optional[List[CarouselMediaItem]] = None


@dataclass
class SharingFrictionInfo(SerializableAttrs):
    should_have_sharing_friction: bool
    bloks_app_url: Optional[str]


# The fields in this class have been observed in reel share items, but may exist elsewhere too.
# If they're observed in other types, they should be moved to MediaShareItem.
@dataclass(kw_only=True)
class ReelMediaShareItem(MediaShareItem, SerializableAttrs):
    # These three are apparently sometimes not present
    # TODO enum?
    caption_position: Optional[int] = None
    is_reel_media: Optional[bool] = None
    timezone_offset: Optional[int] = None
    # likers: List[TODO]
    can_see_insights_as_brand: bool = False
    expiring_at: Optional[int] = None
    sharing_friction_info: Optional[SharingFrictionInfo] = None
    is_in_profile_grid: bool = False
    profile_grid_control_enabled: bool = False
    is_shop_the_look_eligible: bool = False
    # TODO enum?
    deleted_reason: Optional[int] = None
    integrity_review_decision: Optional[str] = None
    # Not present in story_share, only reel_share
    story_is_saved_to_archive: Optional[bool] = None


@dataclass(kw_only=True)
class ReplayableMediaItem(SerializableAttrs):
    view_mode: ViewMode
    seen_count: int
    seen_user_ids: List[int]
    replay_expiring_at_us: Optional[int] = None


@dataclass(kw_only=True)
class VisualMedia(ReplayableMediaItem, SerializableAttrs):
    url_expire_at_secs: Optional[int] = None
    playback_duration_secs: Optional[int] = None
    media: Union[RegularMediaItem, ExpiredMediaItem]

    @classmethod
    def deserialize(cls, data: JSON) -> "VisualMedia":
        data = {**data}
        if "id" not in data["media"]:
            data["media"] = ExpiredMediaItem.deserialize(data["media"])
        else:
            data["media"] = RegularMediaItem.deserialize(data["media"])
        return _dict_to_attrs(cls, data)


@dataclass(kw_only=True)
class AudioInfo(SerializableAttrs):
    audio_src: str
    duration: int
    waveform_data: Optional[List[int]] = None
    waveform_sampling_frequency_hz: Optional[int] = None


@dataclass(kw_only=True)
class VoiceMediaData(SerializableAttrs):
    id: str
    audio: AudioInfo
    organic_tracking_token: str
    user: UserIdentifier
    # TODO enum?
    product_type: str  # "direct_audio"
    media_type: MediaType  # MediaType.AUDIO


@dataclass(kw_only=True)
class VoiceMediaItem(ReplayableMediaItem, SerializableAttrs):
    media: VoiceMediaData


@dataclass(kw_only=True)
class AnimatedMediaImage(SerializableAttrs):
    height: str
    mp4: str
    mp4_size: str
    size: str
    url: str
    webp: str
    webp_size: str
    width: str


@dataclass(kw_only=True)
class AnimatedMediaImages(SerializableAttrs):
    fixed_height: Optional[AnimatedMediaImage] = None


@dataclass(kw_only=True)
class AnimatedMediaItem(SerializableAttrs):
    id: str
    images: AnimatedMediaImages
    # user: {is_verified: bool, username: str}
    # is_random: str | None
    # is_sticker: str | bool


class ReactionType(SerializableEnum):
    LIKES = "likes"
    EMOJIS = "emojis"


@dataclass
class Reaction(SerializableAttrs):
    sender_id: int
    timestamp: Optional[int]
    emoji: Optional[str] = "❤️"
    super_react_type: Optional[str] = None
    client_context: Optional[str] = None
    type: Optional[ReactionType] = None

    @property
    def timestamp_ms(self) -> int:
        return self.timestamp // 1000


@dataclass
class Reactions(SerializableAttrs):
    likes_count: int = 0
    likes: List[Reaction] = attr.ib(factory=lambda: [])
    emojis: List[Reaction] = attr.ib(factory=lambda: [])


@dataclass
class LinkContext(SerializableAttrs):
    link_url: str
    link_title: str
    link_summary: str
    link_image_url: str


@dataclass
class LinkItem(SerializableAttrs):
    text: str
    link_context: LinkContext
    client_context: str
    mutation_token: Optional[str] = None


class ReelShareType(ExtensibleEnum):
    REPLY = "reply"
    REACTION = "reaction"
    MENTION = "mention"
    REPLY_GIF = "reply_gif"


@dataclass
class ReelShareReactionInfo(SerializableAttrs):
    emoji: str
    # TODO find type
    # intensity: Any


@dataclass
class ReelShareItem(SerializableAttrs):
    text: str
    type: ReelShareType
    reel_owner_id: int
    is_reel_persisted: int
    reel_type: str
    media: Union[ReelMediaShareItem, ExpiredMediaItem]
    reaction_info: Optional[ReelShareReactionInfo] = None
    mentioned_user_id: Optional[int] = None

    @classmethod
    def deserialize(cls, data: JSON) -> "ReelShareItem":
        data = {**data}
        if "id" not in data["media"]:
            data["media"] = ExpiredMediaItem.deserialize(data["media"])
        else:
            data["media"] = ReelMediaShareItem.deserialize(data["media"])
        return _dict_to_attrs(cls, data)


@dataclass
class StoryShareItem(SerializableAttrs):
    text: str
    media: Union[ReelMediaShareItem, ExpiredMediaItem]

    # Only present when not expired
    is_reel_persisted: Optional[bool] = None
    # TODO enum?
    reel_type: Optional[str] = None  # user_reel
    reel_id: Optional[str] = None
    # TODO enum?
    story_share_type: Optional[str] = None  # default

    # Only present when expired
    message: Optional[str] = None
    # TODO enum
    reason: Optional[int] = None  # 3 = expired?

    @classmethod
    def deserialize(cls, data: JSON) -> "StoryShareItem":
        data = {**data}
        if "media" not in data:
            data["media"] = ExpiredMediaItem()
        else:
            data["media"] = ReelMediaShareItem.deserialize(data["media"])
        return _dict_to_attrs(cls, data)


@dataclass
class DirectMediaShareItem(SerializableAttrs):
    text: str
    # TODO enum?
    media_share_type: str  # tag
    tagged_user_id: int
    media: MediaShareItem


@dataclass
class XMAMediaShareItem(SerializableAttrs):
    xma_layout_type: int

    target_url: str

    title_text: str
    header_title_text: str
    # subtitle_text: Optional[str]

    preview_url: str
    preview_url_mime_type: str
    preview_width: int
    preview_height: int


@dataclass
class ClipItem(SerializableAttrs):
    # TODO there are some additional fields in clips
    clip: MediaShareItem


@dataclass
class FelixShareItem(SerializableAttrs):
    video: MediaShareItem
    text: Optional[str] = None


@dataclass
class ProfileItem(BaseResponseUser, SerializableAttrs):
    pass


@dataclass(kw_only=True)
class ThreadItem(SerializableAttrs):
    item_id: Optional[str] = None
    user_id: Optional[int] = None
    timestamp: int = 0
    item_type: Optional[ThreadItemType] = None
    is_shh_mode: bool = False
    new_reaction: Optional[Reaction] = None

    text: Optional[str] = None
    client_context: Optional[str] = None
    show_forward_attribution: Optional[bool] = None
    action_log: Optional[ThreadItemActionLog] = None
    auxiliary_text: Optional[str] = None
    auxiliary_text_source_type: Optional[int] = None
    message_item_type: Optional[str] = None

    replied_to_message: Optional["ThreadItem"] = None

    media: Optional[RegularMediaItem] = None
    voice_media: Optional[VoiceMediaItem] = None
    animated_media: Optional[AnimatedMediaItem] = None
    visual_media: Optional[VisualMedia] = None
    media_share: Optional[MediaShareItem] = None
    direct_media_share: Optional[DirectMediaShareItem] = None
    xma_media_share: Optional[List[XMAMediaShareItem]] = None
    xma_story_share: Optional[List[XMAMediaShareItem]] = None
    xma_reel_share: Optional[List[XMAMediaShareItem]] = None
    xma_reel_mention: Optional[List[XMAMediaShareItem]] = None
    reel_share: Optional[ReelShareItem] = None
    story_share: Optional[StoryShareItem] = None
    location: Optional[Location] = None
    reactions: Optional[Reactions] = None
    like: Optional[str] = None
    link: Optional[LinkItem] = None
    clip: Optional[ClipItem] = None
    felix_share: Optional[FelixShareItem] = None
    profile: Optional[ProfileItem] = None

    @property
    def timestamp_ms(self) -> int:
        return self.timestamp // 1000

    @classmethod
    def deserialize(cls, data: JSON, catch_errors: bool = True) -> Union["ThreadItem", Obj]:
        if not catch_errors:
            return _dict_to_attrs(cls, data)
        try:
            return _dict_to_attrs(cls, data)
        except SerializerError:
            log.debug("Failed to deserialize ThreadItem %s", data, exc_info=True)
            return Obj(**data)

    @property
    def unhandleable_type(self) -> str:
        if self.action_log:
            return "action log"
        return "unknown"

    @property
    def is_handleable(self) -> bool:
        return bool(
            self.media
            or self.animated_media
            or self.voice_media
            or self.visual_media
            or self.location
            or self.profile
            or self.reel_share
            or self.media_share
            or self.direct_media_share
            or self.story_share
            or self.clip
            or self.felix_share
            or self.text
            or self.like
            or self.link
        )


# This resolves the 'ThreadItem' string into an actual type.
# Starting Python 3.10, all type annotations will be strings and have to be resolved like this.
# TODO do this automatically for all SerializableAttrs somewhere in mautrix-python
attr.resolve_types(ThreadItem)



# mautrix-instagram - A Matrix-Instagram puppeting bridge.
# Copyright (C) 2020 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Optional, Union

from attr import dataclass

from mautrix.types import SerializableAttrs


@dataclass
class AndroidSession(SerializableAttrs):
    eu_dc_enabled: Optional[bool] = None
    thumbnail_cache_busting_value: int = 1000
    ads_opt_out: bool = False

    ig_www_claim: Optional[str] = None
    authorization: Optional[str] = None
    password_encryption_pubkey: Optional[str] = None
    password_encryption_key_id: Union[None, str, int] = None
    region_hint: Optional[str] = None

    shbid: Optional[str] = None
    shbts: Optional[str] = None
    ds_user_id: Optional[str] = None
    rur: Optional[str] = None



# mautrix-instagram - A Matrix-Instagram puppeting bridge.
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

from typing import Any
import asyncio
import logging

from mautrix.bridge import Bridge
from mautrix.types import RoomID, UserID

from . import commands
from .config import Config
from .db import init as init_db, upgrade_table
from .matrix import MatrixHandler
from .portal import Portal
from .puppet import Puppet
from .user import User
from .version import linkified_version, version
from .web import ProvisioningAPI


class InstagramBridge(Bridge):
    name = "mautrix-instagram"
    module = "mautrix_instagram"
    command = "python -m mautrix-instagram"
    description = "A Matrix-Instagram DM puppeting bridge."
    repo_url = "https://github.com/mautrix/instagram"
    version = version
    markdown_version = linkified_version
    config_class = Config
    matrix_class = MatrixHandler
    upgrade_table = upgrade_table

    config: Config
    matrix: MatrixHandler
    provisioning_api: ProvisioningAPI

    periodic_reconnect_task: asyncio.Task | None

    def preinit(self) -> None:
        self.periodic_reconnect_task = None
        super().preinit()

    def prepare_db(self) -> None:
        super().prepare_db()
        init_db(self.db)

    def prepare_bridge(self) -> None:
        super().prepare_bridge()
        cfg = self.config["bridge.provisioning"]
        self.provisioning_api = ProvisioningAPI(
            shared_secret=cfg["shared_secret"],
            device_seed=self.config["instagram.device_seed"],
            segment_key=cfg["segment_key"],
        )
        self.az.app.add_subapp(cfg["prefix"], self.provisioning_api.app)

    async def start(self) -> None:
        self.add_startup_actions(User.init_cls(self))
        self.add_startup_actions(Puppet.init_cls(self))
        Portal.init_cls(self)
        if self.config["bridge.resend_bridge_info"]:
            self.add_startup_actions(self.resend_bridge_info())
        await super().start()
        self.periodic_reconnect_task = asyncio.create_task(self._try_periodic_reconnect_loop())

    def prepare_stop(self) -> None:
        if self.periodic_reconnect_task is not None and not self.periodic_reconnect_task.done():
            self.periodic_reconnect_task.cancel()
        self.add_shutdown_actions(user.stop_listen() for user in User.by_igpk.values())
        self.log.debug("Stopping puppet syncers")
        for puppet in Puppet.by_custom_mxid.values():
            puppet.stop()

    async def resend_bridge_info(self) -> None:
        self.config["bridge.resend_bridge_info"] = False
        self.config.save()
        self.log.info("Re-sending bridge info state event to all portals")
        async for portal in Portal.all_with_room():
            await portal.update_bridge_info()
        self.log.info("Finished re-sending bridge info state events")

    async def _try_periodic_reconnect_loop(self) -> None:
        try:
            await self._periodic_reconnect_loop()
        except Exception:
            self.log.exception("Fatal error in periodic reconnect loop")

    async def _periodic_reconnect_loop(self) -> None:
        log = logging.getLogger("mau.periodic_reconnect")
        always_reconnect = self.config["bridge.periodic_reconnect.always"]
        interval = self.config["bridge.periodic_reconnect.interval"]
        if interval <= 0:
            log.debug("Periodic reconnection is not enabled")
            return
        resync = bool(self.config["bridge.periodic_reconnect.resync"])
        if interval < 600:
            log.warning("Periodic reconnect interval is quite low (%d)", interval)
        log.debug("Starting periodic reconnect loop")
        while True:
            try:
                await asyncio.sleep(interval)
            except asyncio.CancelledError:
                log.debug("Periodic reconnect loop stopped")
                return
            log.info("Executing periodic reconnections")
            for user in User.by_igpk.values():
                if not user.client or (not user.is_connected and not always_reconnect):
                    log.debug("Not reconnecting %s: not connected", user.mxid)
                    continue
                log.debug("Executing periodic reconnect for %s", user.mxid)
                try:
                    await user.refresh(resync=resync)
                except asyncio.CancelledError:
                    log.debug("Periodic reconnect loop stopped")
                    return
                except Exception:
                    log.exception("Error while reconnecting %s", user.mxid)

    async def get_portal(self, room_id: RoomID) -> Portal:
        return await Portal.get_by_mxid(room_id)

    async def get_puppet(self, user_id: UserID, create: bool = False) -> Puppet:
        return await Puppet.get_by_mxid(user_id, create=create)

    async def get_double_puppet(self, user_id: UserID) -> Puppet:
        return await Puppet.get_by_custom_mxid(user_id)

    async def get_user(self, user_id: UserID, create: bool = True) -> User:
        return await User.get_by_mxid(user_id, create=create)

    def is_bridge_ghost(self, user_id: UserID) -> bool:
        return bool(Puppet.get_id_from_mxid(user_id))

    async def count_logged_in_users(self) -> int:
        return len([user for user in User.by_igpk.values() if user.igpk])

    async def manhole_global_namespace(self, user_id: UserID) -> dict[str, Any]:
        return {
            **await super().manhole_global_namespace(user_id),
            "User": User,
            "Portal": Portal,
            "Puppet": Puppet,
        }


InstagramBridge().run()



# mautrix-instagram - A Matrix-Instagram puppeting bridge.
# Copyright (C) 2020 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from ..types import UserSearchResponse
from .base import BaseAndroidAPI


class UserAPI(BaseAndroidAPI):
    async def search_users(self, query: str, count: int = 30) -> UserSearchResponse:
        req = {
            "timezone_offset": self.state.device.timezone_offset,
            "q": query,
            "count": count,
        }
        return await self.std_http_get(
            "/api/v1/users/search/", query=req, response_type=UserSearchResponse
        )



# mautrix-facebook - A Matrix-Facebook Messenger puppeting bridge.
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

from typing import Type, TypeVar
from contextlib import asynccontextmanager
from urllib.parse import quote, urlparse
import base64
import hashlib
import json
import logging
import pkgutil
import random
import time

from aiohttp import ClientResponse, ClientSession
from aiohttp.client import _RequestContextManager
from yarl import URL
import aiohttp
import zstandard as zstd

from mautrix.types import JSON
from mautrix.util.logging import TraceLogger

from ..proxy import ProxyHandler
from ..state import AndroidState
from ..types import GraphQLMutation, GraphQLQuery
from .errors import GraphQLError, ResponseError, ResponseTypeError, error_class_map, error_code_map

try:
    from aiohttp_socks import ProxyConnector
except ImportError:
    ProxyConnector = None


T = TypeVar("T")


@asynccontextmanager
async def sandboxed_get(url: URL) -> _RequestContextManager:
    async with ClientSession() as sess, sess.get(url) as resp:
        yield resp


zstd_dict = zstd.ZstdCompressionDict(data=pkgutil.get_data("maufbapi.http", "zstd-dict.dat"))
zstd_decomp = zstd.ZstdDecompressor(zstd_dict)


class BaseAndroidAPI:
    a_url = URL("https://api.facebook.com")
    b_url = URL("https://b-api.facebook.com")
    graph_url = URL("https://graph.facebook.com")
    b_graph_url = URL("https://b-graph.facebook.com")
    rupload_url = URL("https://rupload.facebook.com")
    http: ClientSession
    log: TraceLogger

    # Seems to be a per-minute request identifier
    _cid: str
    _cid_ts: int
    freeze_cid: bool
    # Seems to be a per-session request identifier
    nid: str
    # Seems to be a per-request incrementing integer
    _tid: int

    def __init__(
        self,
        state: AndroidState,
        log: TraceLogger | None = None,
        proxy_handler: ProxyHandler | None = None,
    ) -> None:
        self.log = log or logging.getLogger("maufbapi.http")

        self.proxy_handler = proxy_handler
        self.setup_http()

        self.state = state
        self._cid = ""
        self._cid_ts = 0
        self.freeze_cid = False
        self.nid = base64.b64encode(
            bytes([random.getrandbits(8) for _ in range(9)]),
        ).decode("utf-8")
        self._tid = 0
        self._file_url_cache = {}

    @property
    def tid(self) -> int:
        self._tid += 1
        return self._tid

    @property
    def cid(self) -> str:
        new_ts = int(time.time() / 60)
        if not self._cid or (self._cid_ts != new_ts and not self.freeze_cid):
            self._cid_ts = new_ts
            rand = random.Random(f"{self.state.device.uuid}{new_ts}")
            self._cid = bytes([rand.getrandbits(8) for _ in range(16)]).hex()
        return self._cid

    @property
    def session_id(self) -> str:
        return f"nid={self.nid};pid=Main;tid={self.tid};nc=0;fc=0;bc=0,cid={self.cid}"

    def format(self, req: dict[str, str], sign: bool = True, **extra: str) -> str:
        req = dict(sorted(req.items()))
        if sign:
            sig_data = "".join(f"{key}={value}" for key, value in req.items())
            sig_data_bytes = (sig_data + self.state.application.client_secret).encode("utf-8")
            req["sig"] = hashlib.md5(sig_data_bytes).hexdigest()
        if extra:
            req.update(extra)
        return "&".join(f"{quote(key)}={quote(value)}" for key, value in sorted(req.items()))

    @property
    def _headers(self) -> dict[str, str]:
        headers = {
            "x-fb-connection-quality": self.state.device.connection_quality,
            "x-fb-connection-type": self.state.device.connection_type,
            "user-agent": self.state.user_agent,
            "x-tigon-is-retry": "False",
            "x-fb-http-engine": "Liger",
            "x-fb-client-ip": "True",
            "x-fb-server-cluster": "True",
            # "x-fb-connection-token": self.cid,
            # "x-fb-session-id": self.session_id,
            "x-fb-device-group": self.state.device.device_group,
            "x-fb-sim-hni": str(self.state.carrier.hni),
            "x-fb-net-hni": str(self.state.carrier.hni),
            "x-fb-rmd": "cached=0;state=NO_MATCH",
            "x-fb-request-analytics-tags": "unknown",
            # "x-fb-background-state": "1",
            "authorization": f"OAuth {self.state.session.access_token or 'null'}",
        }
        return {k: v for k, v in headers.items() if v is not None}

    @property
    def _params(self) -> dict[str, str]:
        return {
            "locale": self.state.device.language,
            "client_country_code": self.state.device.country_code,
        }

    def setup_http(self) -> None:
        connector = None
        http_proxy = self.proxy_handler.get_proxy_url()
        if http_proxy:
            if ProxyConnector:
                connector = ProxyConnector.from_url(http_proxy)
            else:
                self.log.warning("http_proxy is set, but aiohttp-socks is not installed")

        self.http = ClientSession(connector=connector)
        return None

    def get(
        self,
        url: str | URL,
        headers: dict[str, str] | None = None,
        include_auth: bool = True,
        sandbox: bool = False,
        **kwargs,
    ) -> _RequestContextManager:
        headers = {
            **self._headers,
            **(headers or {}),
        }
        orig_url = url
        url = URL(url)
        if "/" in url.query_string:
            # Hacky hacks for forcing encoded slashes in query parameters. Normally yarl/aiohttp
            # forces decoding slashes in query parameters, but that breaks Facebook's URL signature
            # and the CDN rejects the request. (the reason there are slashes in the URL in the
            # first place is presumably that Facebook hasn't heard of URL-safe base64).
            urlparsed = urlparse(orig_url)
            url = url.with_query(None).with_path(
                f"{urlparsed.path}?{urlparsed.query}", encoded=True
            )
        if not url.host.endswith(".facebook.com") or not include_auth:
            headers.pop("authorization")
            if sandbox:
                return sandboxed_get(url)
        return self.http.get(url, headers=headers, **kwargs)

    async def graphql(
        self,
        req: GraphQLQuery,
        headers: dict[str, str] | None = None,
        response_type: Type[T] | None = JSON,
        path: list[str] | None = None,
        b: bool = True,
    ) -> T:
        headers = {
            **self._headers,
            **(headers or {}),
            "content-type": "application/x-www-form-urlencoded",
            "x-fb-friendly-name": req.__class__.__name__,
            "x-fb-request-analytics-tags": "graphservice",
            "accept-encoding": "x-fb-dz;d=1, gzip, deflate",
        }
        variables = req.serialize()
        if isinstance(req, GraphQLMutation):
            variables = {"input": variables}
        params = {
            **self._params,
            "variables": json.dumps(variables),
            "method": "post",
            "doc_id": req.doc_id,
            "format": "json",
            "pretty": "false",
            # "query_name": req.__class__.__name__,
            "strip_defaults": "false",
            "strip_nulls": "false",
            "fb_api_req_friendly_name": req.__class__.__name__,
            "fb_api_caller_class": req.caller_class,
            "fb_api_analytics_tags": json.dumps(req.analytics_tags),
            "server_timestamps": "true",
        }
        if not req.include_client_country_code:
            params.pop("client_country_code")
        resp = await self.http.post(
            url=(self.b_graph_url if b else self.graph_url) / "graphql",
            data=params,
            headers=headers,
        )
        await self._decompress_zstd(resp)
        self.log.trace(f"GraphQL {req} response: {await resp.text()}")
        if response_type is None:
            self._handle_response_headers(resp)
            return None
        json_data = await self._handle_response(resp)
        if path:
            for item in path:
                json_data = json_data[item]
        if response_type is not JSON:
            return response_type.deserialize(json_data)
        return json_data

    async def _decompress_zstd(self, resp: ClientResponse) -> None:
        if (
            resp.headers.get("content-encoding") == "x-fb-dz"
            and resp.headers.get("x-fb-dz-dict") == "1"
            and not getattr(resp, "_zstd_decompressed", None)
        ):
            compressed = await resp.read()
            resp._body = zstd_decomp.decompress(compressed)
            self.log.trace(
                f"Decompressed {len(compressed)} bytes of zstd "
                f"into {len(resp._body)} bytes of (hopefully) JSON"
            )
            setattr(resp, "_zstd_decompressed", True)

    async def _handle_response(self, resp: ClientResponse, batch_index: int | None = None) -> JSON:
        await self._decompress_zstd(resp)
        self._handle_response_headers(resp)
        try:
            body = await resp.json()
        except (json.JSONDecodeError, aiohttp.ContentTypeError) as e:
            raise ResponseTypeError(resp.status, await resp.text()) from e
        if isinstance(body, list) and batch_index is not None:
            body = body[batch_index][1].get("body", {})
        error = body.get("error", None)
        errors = body.get("errors", [])
        if error:
            self.log.trace("Got error object in response data: %s", error)
            error_class = (
                error_code_map.get(error["code"])
                or error_class_map.get(error["type"])
                or ResponseError
            )
            raise error_class(error)
        elif errors:
            self.log.warning("Got list of errors in response data: %s", errors)
            if resp.status >= 400 or not body.get("data"):
                try:
                    raise GraphQLError(errors[0], errors[1:])
                except KeyError as e:
                    raise Exception("Unknown response error") from e
        return body

    def _handle_response_headers(self, resp: ClientResponse) -> None:
        # TODO if needed
        pass



from .base import IGError
from .mqtt import (
    IGMQTTError,
    IrisSubscribeError,
    MQTTConnectionUnauthorized,
    MQTTNotConnected,
    MQTTNotLoggedIn,
)
from .response import (
    IGActionSpamError,
    IGBad2FACodeError,
    IGChallengeError,
    IGChallengeWrongCodeError,
    IGCheckpointError,
    IGConsentRequiredError,
    IGFBNoContactPointFoundError,
    IGInactiveUserError,
    IGLoginBadPasswordError,
    IGLoginError,
    IGLoginInvalidUserError,
    IGLoginRequiredError,
    IGLoginTwoFactorRequiredError,
    IGLoginUnusablePasswordError,
    IGNotFoundError,
    IGNotLoggedInError,
    IGPrivateUserError,
    IGRateLimitError,
    IGResponseError,
    IGSentryBlockError,
    IGUnknownError,
    IGUserHasLoggedOutError,
)
from .state import IGCookieNotFoundError, IGNoChallengeError, IGUserIDNotFoundError



from .account import AccountAPI
from .challenge import ChallengeAPI
from .login import LoginAPI
from .qe import QeSyncAPI
from .thread import ThreadAPI
from .upload import UploadAPI
from .user import UserAPI


class AndroidAPI(ThreadAPI, AccountAPI, QeSyncAPI, LoginAPI, UploadAPI, ChallengeAPI, UserAPI):
    pass



# mautrix-instagram - A Matrix-Instagram puppeting bridge.
# Copyright (C) 2020 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Optional, get_type_hints

from aiohttp import ClientResponse

from mautrix.types import JSON, Serializable

from ..types import (
    ChallengeResponse,
    CheckpointResponse,
    ConsentRequiredResponse,
    LoginErrorResponse,
    LoginRequiredResponse,
    SpamResponse,
)
from .base import IGError


class IGChallengeWrongCodeError(IGError):
    pass


class IGUnknownError(IGError):
    response: ClientResponse

    def __init__(self, response: ClientResponse) -> None:
        super().__init__(
            f"Request {response.request_info.method} {response.request_info.url.path} failed: "
            f"HTTP {response.status} with non-JSON body"
        )
        self.response = response


class IGResponseError(IGError):
    response: ClientResponse

    def __init__(self, response: ClientResponse, json: JSON) -> None:
        prefix = f"Request {response.request_info.method} {response.request_info.url.path} failed"
        message = f"HTTP {response.status}"
        self.response = response
        if "message" in json:
            message = json["message"]
        type_hint = get_type_hints(type(self)).get("body", JSON)
        if type_hint is not JSON and issubclass(type_hint, Serializable):
            self.body = type_hint.deserialize(json)
        else:
            self.body = json
        super().__init__(f"{prefix}: {self._message_override or message}")

    @property
    def _message_override(self) -> Optional[str]:
        return None


class IGActionSpamError(IGResponseError):
    body: SpamResponse

    @property
    def _message(self) -> str:
        return f"HTTP {self.body.message}"


class IGNotFoundError(IGResponseError):
    pass


class IGRateLimitError(IGResponseError):
    pass


class IGCheckpointError(IGResponseError):
    body: CheckpointResponse


class IGChallengeError(IGResponseError):
    body: ChallengeResponse

    @property
    def url(self) -> str:
        return self.body.challenge.api_path


class IGConsentRequiredError(IGResponseError):
    body: ConsentRequiredResponse


class IGNotLoggedInError(IGResponseError):
    body: LoginRequiredResponse

    @property
    def proper_message(self) -> str:
        return (
            f"{self.body.error_title or self.body.message} "
            f"(reason code: {self.body.logout_reason})"
        )


class IGUserHasLoggedOutError(IGNotLoggedInError):
    pass


class IGLoginRequiredError(IGNotLoggedInError):
    pass


class IGPrivateUserError(IGResponseError):
    pass


class IGSentryBlockError(IGResponseError):
    pass


class IGInactiveUserError(IGResponseError):
    pass


class IGLoginError(IGResponseError):
    body: LoginErrorResponse


class IGLoginTwoFactorRequiredError(IGLoginError):
    pass


class IGLoginBadPasswordError(IGLoginError):
    pass


class IGLoginUnusablePasswordError(IGLoginError):
    pass


class IGLoginInvalidUserError(IGLoginError):
    pass


class IGFBNoContactPointFoundError(IGLoginError):
    pass


class IGBad2FACodeError(IGResponseError):
    pass



# mautrix-instagram - A Matrix-Instagram puppeting bridge.
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

from typing import Any, Type, TypeVar
import json
import logging
import random
import time

from aiohttp import ClientResponse, ClientSession, ContentTypeError
from yarl import URL

from mautrix.types import JSON, Serializable
from mautrix.util.logging import TraceLogger

from ..errors import (
    IGActionSpamError,
    IGBad2FACodeError,
    IGChallengeError,
    IGCheckpointError,
    IGConsentRequiredError,
    IGFBNoContactPointFoundError,
    IGInactiveUserError,
    IGLoginBadPasswordError,
    IGLoginInvalidUserError,
    IGLoginRequiredError,
    IGLoginTwoFactorRequiredError,
    IGLoginUnusablePasswordError,
    IGNotFoundError,
    IGPrivateUserError,
    IGRateLimitError,
    IGResponseError,
    IGSentryBlockError,
    IGUnknownError,
    IGUserHasLoggedOutError,
)
from ..state import AndroidState

T = TypeVar("T")


def remove_nulls(d: dict) -> dict:
    return {
        k: remove_nulls(v) if isinstance(v, dict) else v for k, v in d.items() if v is not None
    }


class BaseAndroidAPI:
    url = URL("https://i.instagram.com")
    http: ClientSession
    state: AndroidState
    log: TraceLogger

    def __init__(self, state: AndroidState, log: TraceLogger | None = None) -> None:
        self.http = ClientSession(cookie_jar=state.cookies.jar)
        self.state = state
        self.log = log or logging.getLogger("mauigpapi.http")

    @staticmethod
    def sign(req: Any, filter_nulls: bool = False) -> dict[str, str]:
        if isinstance(req, Serializable):
            req = req.serialize()
        if isinstance(req, dict):
            req = json.dumps(remove_nulls(req) if filter_nulls else req)
        return {"signed_body": f"SIGNATURE.{req}"}

    @property
    def _headers(self) -> dict[str, str]:
        headers = {
            "x-ads-opt-out": str(int(self.state.session.ads_opt_out)),
            "x-device-id": self.state.device.uuid,
            "x-ig-app-locale": self.state.device.language,
            "x-ig-device-locale": self.state.device.language,
            "x-pigeon-session-id": self.state.pigeon_session_id,
            "x-pigeon-rawclienttime": str(round(time.time(), 3)),
            "x-ig-connection-speed": f"{random.randint(1000, 3700)}kbps",
            "x-ig-bandwidth-speed-kbps": "-1.000",
            "x-ig-bandwidth-totalbytes-b": "0",
            "x-ig-bandwidth-totaltime-ms": "0",
            "x-ig-eu-dc-enabled": (
                str(self.state.session.eu_dc_enabled).lower()
                if self.state.session.eu_dc_enabled is not None
                else None
            ),
            "x-ig-app-startup-country": self.state.device.language.split("_")[1],
            "x-bloks-version-id": self.state.application.BLOKS_VERSION_ID,
            "x-ig-www-claim": self.state.session.ig_www_claim or "0",
            "x-bloks-is-layout-rtl": str(self.state.device.is_layout_rtl).lower(),
            "x-bloks-is-panorama-enabled": "true",
            "x-ig-timezone-offset": self.state.device.timezone_offset,
            # "x-messenger": "1",
            "x-ig-device-id": self.state.device.uuid,
            "x-ig-android-id": self.state.device.id,
            "x-ig-connection-type": self.state.device.connection_type,
            "x-ig-capabilities": self.state.application.CAPABILITIES,
            "x-ig-app-id": self.state.application.FACEBOOK_ANALYTICS_APPLICATION_ID,
            "user-agent": self.state.user_agent,
            "accept-language": self.state.device.language.replace("_", "-"),
            "authorization": self.state.session.authorization,
            "x-mid": self.state.cookies.get_value("mid"),
            "ig-u-ig-direct-region-hint": self.state.session.region_hint,
            "ig-u-shbid": self.state.session.shbid,
            "ig-u-shbts": self.state.session.shbts,
            "ig-u-ds-user-id": self.state.session.ds_user_id,
            "ig-u-rur": self.state.session.rur,
            "x-fb-http-engine": "Liger",
            "x-fb-client-ip": "True",
            "accept-encoding": "gzip",
        }
        return {k: v for k, v in headers.items() if v is not None}

    def raw_http_get(self, url: URL | str):
        if isinstance(url, str):
            url = URL(url, encoded=True)
        return self.http.get(
            url,
            headers={
                "user-agent": self.state.user_agent,
                "accept-language": self.state.device.language.replace("_", "-"),
            },
        )

    async def std_http_post(
        self,
        path: str,
        data: JSON = None,
        raw: bool = False,
        filter_nulls: bool = False,
        headers: dict[str, str] | None = None,
        query: dict[str, str] | None = None,
        response_type: Type[T] | None = JSON,
    ) -> T:
        headers = {**self._headers, **headers} if headers else self._headers
        if not raw:
            data = self.sign(data, filter_nulls=filter_nulls)
        url = self.url.with_path(path).with_query(query or {})
        resp = await self.http.post(url=url, headers=headers, data=data)
        self.log.trace(f"{path} response: {await resp.text()}")
        if response_type is str or response_type is None:
            self._handle_response_headers(resp)
            if response_type is str:
                return await resp.text()
            return None
        json_data = await self._handle_response(resp)
        if response_type is not JSON:
            return response_type.deserialize(json_data)
        return json_data

    async def std_http_get(
        self,
        path: str,
        query: dict[str, str] | None = None,
        headers: dict[str, str] | None = None,
        response_type: Type[T] | None = JSON,
    ) -> T:
        headers = {**self._headers, **headers} if headers else self._headers
        query = {k: v for k, v in (query or {}).items() if v is not None}
        resp = await self.http.get(url=self.url.with_path(path).with_query(query), headers=headers)
        self.log.trace(f"{path} response: {await resp.text()}")
        if response_type is None:
            self._handle_response_headers(resp)
            return None
        json_data = await self._handle_response(resp)
        if response_type is not JSON:
            return response_type.deserialize(json_data)
        return json_data

    async def _handle_response(self, resp: ClientResponse) -> JSON:
        self._handle_response_headers(resp)
        try:
            body = await resp.json()
        except (json.JSONDecodeError, ContentTypeError) as e:
            raise IGUnknownError(resp) from e
        if body.get("status", "fail") == "ok":
            return body
        else:
            await self._raise_response_error(resp)

    async def _raise_response_error(self, resp: ClientResponse) -> None:
        try:
            data = await resp.json()
        except json.JSONDecodeError:
            data = {}

        if data.get("spam", False):
            raise IGActionSpamError(resp, data)
        elif data.get("two_factor_required", False):
            raise IGLoginTwoFactorRequiredError(resp, data)
        elif resp.status == 404:
            raise IGNotFoundError(resp, data)
        elif resp.status == 429:
            raise IGRateLimitError(resp, data)

        message = data.get("message")
        if isinstance(message, str):
            if message == "challenge_required":
                err = IGChallengeError(resp, data)
                self.state.challenge_path = err.url
                raise err
            elif message == "checkpoint_required":
                raise IGCheckpointError(resp, data)
            elif message == "consent_required":
                raise IGConsentRequiredError(resp, data)
            elif message == "user_has_logged_out":
                raise IGUserHasLoggedOutError(resp, data)
            elif message == "login_required":
                raise IGLoginRequiredError(resp, data)
            elif message.lower() == "not authorized to view user":
                raise IGPrivateUserError(resp, data)

        error_type = data.get("error_type")
        if error_type == "sentry_block":
            raise IGSentryBlockError(resp, data)
        elif error_type == "inactive_user":
            raise IGInactiveUserError(resp, data)
        elif error_type == "bad_password":
            raise IGLoginBadPasswordError(resp, data)
        elif error_type == "unusable_password":
            raise IGLoginUnusablePasswordError(resp, data)
        elif error_type == "invalid_user":
            raise IGLoginInvalidUserError(resp, data)
        elif error_type == "sms_code_validation_code_invalid":
            raise IGBad2FACodeError(resp, data)
        elif error_type == "fb_no_contact_point_found":
            raise IGFBNoContactPointFoundError(resp, data)

        raise IGResponseError(resp, data)

    def _handle_response_headers(self, resp: ClientResponse) -> None:
        fields = {
            "x-ig-set-www-claim": "ig_www_claim",
            "ig-set-authorization": "authorization",
            "ig-set-password-encryption-key-id": "password_encryption_key_id",
            "ig-set-password-encryption-pub-key": "password_encryption_pubkey",
            "ig-set-ig-u-ig-direct-region-hint": "region_hint",
            "ig-set-ig-u-shbid": "shbid",
            "ig-set-ig-u-shbts": "shbts",
            "ig-set-ig-u-rur": "rur",
            "ig-set-ig-u-ds-user-id": "ds_user_id",
        }
        for header, field in fields.items():
            value = resp.headers.get(header)
            if value and (header != "IG-Set-Authorization" or not value.endswith(":")):
                setattr(self.state.session, field, value)



# mautrix-instagram - A Matrix-Instagram puppeting bridge.
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

from uuid import uuid4
import json
import random
import time

from ..types import FinishUploadResponse, MediaType, UploadPhotoResponse, UploadVideoResponse
from .base import BaseAndroidAPI


class UploadAPI(BaseAndroidAPI):
    async def upload_photo(
        self,
        data: bytes,
        mime: str,
        upload_id: str | None = None,
        width: int | None = None,
        height: int | None = None,
    ) -> UploadPhotoResponse:
        upload_id = upload_id or str(int(time.time() * 1000))
        name = f"{upload_id}_0_{random.randint(1000000000, 9999999999)}"
        params = {
            "retry_context": json.dumps(
                {
                    "num_step_auto_retry": 0,
                    "num_reupload": 0,
                    "num_step_manual_retry": 0,
                }
            ),
            "media_type": str(MediaType.IMAGE.value),
            "upload_id": upload_id,
            "xsharing_user_ids": json.dumps([]),
        }
        if mime == "image/jpeg":
            params["image_compression"] = json.dumps(
                {"lib_name": "moz", "lib_version": "3.1.m", "quality": 80}
            )
        if width and height:
            params["original_width"] = str(width)
            params["original_height"] = str(height)
        headers = {
            "X_FB_PHOTO_WATERFALL_ID": str(uuid4()),
            "X-Entity-Type": mime,
            "Offset": "0",
            "X-Instagram-Rupload-Params": json.dumps(params),
            "X-Entity-Name": name,
            "X-Entity-Length": str(len(data)),
            "Content-Type": "application/octet-stream",
            "priority": "u=6, i",
        }
        return await self.std_http_post(
            f"/rupload_igphoto/{name}",
            headers=headers,
            data=data,
            raw=True,
            response_type=UploadPhotoResponse,
        )

    async def upload_mp4(
        self,
        data: bytes,
        upload_id: str | None = None,
        audio: bool = False,
        duration_ms: int | None = None,
        width: int | None = None,
        height: int | None = None,
    ) -> tuple[UploadVideoResponse, str]:
        upload_id = upload_id or str(int(time.time() * 1000))
        name = f"{upload_id}_0_{random.randint(1000000000, 9999999999)}"
        media_type = MediaType.AUDIO if audio else MediaType.VIDEO
        params: dict[str, str] = {
            "retry_context": json.dumps(
                {
                    "num_step_auto_retry": 0,
                    "num_reupload": 0,
                    "num_step_manual_retry": 0,
                }
            ),
            "media_type": str(media_type.value),
            "upload_id": upload_id,
            "xsharing_user_ids": json.dumps([]),
        }
        if duration_ms:
            params["upload_media_duration_ms"] = str(duration_ms)
        if audio:
            params["is_direct_voice"] = "1"
        else:
            params["direct_v2"] = "1"
            params["for_direct_story"] = "1"
            params["content_tags"] = "use_default_cover"
            params["extract_cover_frame"] = "1"
            if width and height:
                params["upload_media_width"] = str(width)
                params["upload_media_height"] = str(height)
        headers = {
            "X_FB_VIDEO_WATERFALL_ID": str(uuid4()),
            "X-Entity-Type": "audio/mp4" if audio else "video/mp4",
            "Offset": "0",
            "X-Instagram-Rupload-Params": json.dumps(params),
            "X-Entity-Name": name,
            "X-Entity-Length": str(len(data)),
            "Content-Type": "application/octet-stream",
            "priority": "u=6, i",
        }
        if not audio:
            headers["segment-type"] = "3"
            headers["segment-start-offset"] = "0"
        return (
            await self.std_http_post(
                f"/rupload_igvideo/{name}",
                headers=headers,
                data=data,
                raw=True,
                response_type=UploadVideoResponse,
            ),
            upload_id,
        )

    async def finish_upload(
        self, upload_id: str, source_type: str, video: bool = False
    ) -> FinishUploadResponse:
        headers = {
            "retry_context": json.dumps(
                {
                    "num_step_auto_retry": 0,
                    "num_reupload": 0,
                    "num_step_manual_retry": 0,
                }
            ),
        }
        req = {
            "timezone_offset": self.state.device.timezone_offset,
            "_csrftoken": self.state.cookies.csrf_token,
            "source_type": source_type,
            "_uid": self.state.session.ds_user_id,
            "device_id": self.state.device.id,
            "_uuid": self.state.device.uuid,
            "upload_id": upload_id,
            "device": self.state.device.payload,
        }
        query = {}
        if video:
            query["video"] = "1"
        return await self.std_http_post(
            "/api/v1/media/upload_finish/",
            headers=headers,
            data=req,
            query=query,
            response_type=FinishUploadResponse,
        )



# mautrix-facebook - A Matrix-Facebook Messenger puppeting bridge.
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

from urllib.parse import urlencode
import json

from ..types.graphql import LoggedInUser
from ..types.graphql.queries import NTContext
from .base import BaseAndroidAPI


class PostLoginAPI(BaseAndroidAPI):
    async def fetch_logged_in_user(self, post_login: bool = False) -> LoggedInUser | None:
        url = self.b_graph_url.with_query(
            {
                "include_headers": "false",
                "decode_body_json": "false",
                "streamable_json_response": "true",
                **self._params,
            }
        )
        req_data = (
            {
                "fb_api_caller_class": "LoginOperations",
                "fb_api_req_friendly_name": "handleLogin",
                "batch": json.dumps(self._post_login_params, separators=(",", ":")),
            }
            if post_login
            else {
                "fb_api_caller_class": "MessagesSyncLoggedInUserFetcher",
                "fb_api_req_friendly_name": "syncRefetchLoggedInUser",
                "batch": json.dumps(self._resync_params, separators=(",", ":")),
            }
        )
        headers = {
            "x-zero-state": "unknown",
            "x-fb-request-analytics-tags": "unknown",
            "x-fb-friendly-name": req_data["fb_api_req_friendly_name"],
            "accept-encoding": "x-fb-dz;d=1, gzip, deflate",
            **self._headers,
        }
        headers.pop("x-fb-rmd", None)
        resp = await self.http.post(url=url, headers=headers, data=req_data)
        await self._decompress_zstd(resp)
        self.log.trace(f"Fetch logged in user response: {await resp.text()}")
        resp_data = await self._handle_response(resp, batch_index=2 if post_login else 0)
        if not post_login:
            # The second batch will sometimes contain errors that the first one doesn't.
            await self._handle_response(resp, batch_index=1)
        try:
            actual_data = resp_data["data"]["viewer"]["actor"]
        except (IndexError, KeyError):
            self.log.warning(
                "Didn't get expected data in fetch logged in user response: %s", resp_data
            )
            return None
        info = LoggedInUser.deserialize(actual_data)
        info.unrecognized_ = {}
        return info

    # fmt: off

    @property
    def _resync_params(self):
        return [
            {
                "method": "POST",
                "body": urlencode({
                    "variables": json.dumps({
                        "profile_pic_small_size": 110,
                        "profile_pic_medium_size": 258,
                        "profile_pic_large_size": 1080,
                        "is_for_messenger": True,
                        "fetch_story_holdout": False,
                    }, separators=(",", ":")),
                    "method": "post",
                    "doc_id": "4043011582467507",
                    "query_name": "GetLoggedInUserQuery",
                    "strip_defaults": "true",
                    "strip_nulls": "true",
                    "locale": self.state.device.language,
                    "client_country_code": self.state.device.country_code,
                    "fb_api_req_friendly_name": "GetLoggedInUserQuery",
                }),
                "name": "user",
                "omit_response_on_success": False,
                "relative_url": "graphql",
            },
            {
                "method": "POST",
                "body": urlencode({
                    "method": "post",
                    "doc_id": "5263028057057401",
                    "query_name": "FetchFacebookEmployeeStatusQuery",
                    "strip_defaults": "true",
                    "strip_nulls": "true",
                    "locale": self.state.device.language,
                    "client_country_code": self.state.device.country_code,
                    "fb_api_req_friendly_name": "FetchFacebookEmployeeStatusQuery",
                }),
                "name": "fetchFacebookEmployeeStatus",
                "omit_response_on_success": False,
                "relative_url": "graphql",
            },
        ]

    @property
    def _post_login_params(self):
        fetch_zero_token_params = {
            "carrier_mcc": self.state.carrier.mcc,
            "carrier_mnc": self.state.carrier.mnc,
            "sim_mcc": self.state.carrier.mcc,
            "sim_mnc": self.state.carrier.mnc,
            "format": "json",
            "interface": self.state.device.connection_type.lower(),
            "dialtone_enabled": "false",
            "needs_backup_rules": "true",
            "request_reason": "login",
            "locale": self.state.device.language,
            "client_country_code": self.state.device.country_code,
            "fb_api_req_friendly_name": "fetchZeroToken",
        }
        return [
            {
                "method": "POST",
                "body": urlencode(fetch_zero_token_params),
                "name": "fetchZeroToken",
                "omit_response_on_success": False,
                "relative_url": "mobile_zero_campaign",
            },
            {
                "method": "POST",
                "body": urlencode({**fetch_zero_token_params, "dialtone_enabled": "true"}),
                "name": "fetchZeroTokenForDialtone",
                "omit_response_on_success": False,
                "relative_url": "mobile_zero_campaign",
            },
            self._resync_params[0],
            {
                "method": "POST",
                "body": urlencode({
                    "variables": json.dumps({
                        "nux_ids": [
                            "9151", "9580", "8587", "5551", "5477", "7441", "6199", "8409", "8330",
                            "6504", "7785", "8435", "6671", "8166", "8071", "10187", "10018",
                            "9682", "10010", "9717", "8432", "8460", "8538", "4442", "9871",
                            "9877", "10186", "9582", "10087", "9890", "7685", "7093", "7841",
                            "6778", "8588", "8266", "9849", "9485", "9455", "9790", "10112",
                            "6849", "3931", "9891", "7721", "7066", "9821", "6762", "5665", "5677",
                            "9606", "9856", "7859", "7826", "7764", "7951", "8478", "7830", "7829",
                            "10003", "6522", "6563", "6005", "7842", "6006", "7979", "7915",
                            "9956", "8449", "5290", "5291", "4828", "7004", "7615", "4745", "4744",
                            "4743", "2415", "4408", "8470", "3545", "5579", "3543", "5411", "8729",
                            "9045", "9808", "9785", "7896", "9608", "8216", "8279", "9423", "7198",
                            "7199", "7190", "7654", "9172", "4541", "9280", "8026", "7987", "8214",
                            "7435", "1820", "9054", "8375", "8543", "4320", "8311", "6389", "4757",
                            "9807", "5131", "4327"
                        ],
                        "device_id": self.state.device.uuid,
                        "nt_context": NTContext().serialize(),
                        "avatar_nux_image_width": 1080,
                        "is_from_internal_tool": False,
                        "family_device_id": self.state.device.fdid,
                        "scale": "3",
                    }, separators=(",", ":")),
                    "method": "post",
                    "doc_id": "4618754314834440",
                    "query_name": "FetchInterstitials",
                    "strip_defaults": "true",
                    "strip_nulls": "true",
                    "locale": self.state.device.language,
                    "client_country_code": self.state.device.country_code,
                    "fb_api_req_friendly_name": "FetchInterstitials"
                }),
                "name": "fetch_interstititals_graphql",
                "omit_response_on_success": False,
                "relative_url": "graphql",
            },
            {
                "method": "POST",
                "body": urlencode({
                    "client_country_code": self.state.device.country_code,
                    "fb_api_req_friendly_name": "fetchGKInfo",
                    "format": "json",
                    "locale": self.state.device.language,
                    "query": "android_aborthooks_enabled,android_actionbar_panel_kb_workaround,android_allow_user_cert_override,android_always_play_live_cached_data,android_analytics_listener_bg_handler,android_block_drm_playback_hdmi,android_block_drm_screen_capture,android_chat_head_hw_accel_disabled,android_chat_heads_app_state,android_cm_holiday_card_render,android_disable_messenger_ineedinits,android_drm_blacklisted_devices,android_enable_maxwidth_prefilter,android_enable_oxygen_crash_reporter,android_enable_terminate_handler,android_enable_vod_prefetch_qs_fix,android_fb4a_enable_zero_ip_test,android_fbns_dumpsys_ids,android_game_tab_badging,android_granular_exposures_navigation,android_headspin_logging,android_invite_link_phone_confirm_enable_gk,android_large_payload_support,android_learn_startvideosession_uri,android_legacy_logging_framework_gk,android_litho_enable_thread_tracing_stacktrace,android_litho_opt_visibility_handlers,android_litho_transitions_extension_290,android_loaded_library_reporting,android_messenger_add_contacts_redesign_gk,android_messenger_android_auto_support,android_messenger_avoid_ipc_logging,android_messenger_background_contact_logs_upload,android_messenger_background_contacts_upload,android_messenger_banner_triggers_omnistore,android_messenger_bug_report_on_instacrash_loop,android_messenger_camera_selfie_effect_gk,android_messenger_composer_lightweight_actions,android_messenger_contact_upload_progress_screens,android_messenger_dark_mode_opt_in,android_messenger_dark_mode_rollout,android_messenger_data_saver_mode,android_messenger_delay_analytics_during_startup,android_messenger_delay_periodic_reporters,android_messenger_employee_default_beta_web_tier,android_messenger_explicitly_adding_extensions,android_messenger_file_based_logger,android_messenger_hole_check_disabled,android_messenger_inapp_browser,android_messenger_instrumented_drawable,android_messenger_log_messaging_debug_events,android_messenger_multiple_requests_money_fab,android_messenger_omnistore_debug_uploader_flytrap,android_messenger_omnistore_rage_shake_sqlite,android_messenger_p2p_payment,android_messenger_periodic_effect_prefetch_enabled,android_messenger_platform_20141218,android_messenger_platform_20150311,android_messenger_platform_20150314,android_messenger_post_capture_effect_gk,android_messenger_privacy_aware_logger,android_messenger_profile_view_redirect_target,android_messenger_rage_shake_enable_loom,android_messenger_rage_shake_in_chat_heads,android_messenger_recipient_holdout_filter,android_messenger_reddit_link,android_messenger_send_or_request_money_fab,android_messenger_show_onboarding_flow,android_messenger_skip_messages_update_on_paused,android_messenger_skip_startup_prefetch,android_messenger_sms_invites_gk,android_messenger_store_in_private_storage,android_mobile_config_sampled_access_disabled,android_mqtt_fast_send,android_mqtt_log_time,android_mqtt_new_wake_lock,android_mqtt_pendingmessage_connect,android_mqtt_reconnect_back_off_fix,android_mqtt_report_connect_sent_state,android_mqttlite_health_stats_sampling,android_mqttlite_log_sampling,android_omnistore_init_using_critical_path_task,android_professional_services_booking,android_qpl_use_freemode_stats,android_qpl_use_image_stats,android_qpl_use_mobile_infra_memory_stats,android_qpl_use_mobileboost_usage,android_qpl_use_msys_info,android_qpl_use_navigation_data,android_qpl_use_nt_stats,android_qpl_use_sapienz_data,android_qpl_use_thermal_stats,android_qpl_use_traffic_transport_monitor_metadata,android_qpl_use_user_perceptible_scopes,android_qpl_use_yoga_provider,android_reliability_lacrima_gk,android_rtc_msqrd_supported,android_run_gc_on_trim,android_show_whitehat_settings,android_soft_error_disabled,android_soft_error_write_to_qpl,android_trusted_tester,android_two_phase_video_spinner,android_video_abr_instrumentation_sampling,android_video_cache_refresh_rate,android_video_delayed_vps_session_release,android_video_detect_plugin_multi_container,android_video_enable_adaptive_caption,android_video_fallback_when_no_reps,android_video_fix_social_player_reload_same_video,android_video_hash_url,android_video_init_constraints_with_highest_format,android_video_live_current_null_as_low_buffer,android_video_live_trace,android_video_live_use_contextual_parameters,android_video_log_stall_detail_event,android_video_playback_getserializable_blacklist,android_video_prefetch_when_no_reps,android_video_profiler_marauder_logging,android_video_refresh_expired_url,android_video_report_prefetch_abr_decision,android_video_resolve_cc_per_process,android_video_resolve_cc_per_video,android_video_rvp_deprecation_notice,android_video_scale_tv_at_ui_thread,android_video_send_debug_headers_to_cdn,android_video_skip_bug_report_extra,android_video_skip_texture_view_get_bitmap,android_video_treat_current_null_as_low_buffer,android_video_use_contextual_network_aware_params,android_video_use_contextual_parameters,android_video_wall_time_protect,android_video_warm_codec,android_whistle_liger_merge,android_whistle_proxy_support,android_whitehat_setting_intercept_traffic,android_zero_optin_graphql_fetch,android_zero_rating_header_request,app_module_download,boost_counters_activity_thread,boost_counters_binder,boost_counters_blockidlejobs,boost_counters_cpuboost,boost_counters_delayedanalytics,boost_counters_graphql,boost_counters_io_thread_periodic,boost_counters_litho_layout_thread,boost_counters_renderthreadboost,boost_counters_smart_fsync,boost_counters_smartgc,boost_counters_softkeyboard,boost_counters_threadaffinity,boost_counters_ui_thread,boost_counters_ui_thread_periodic,campaign_api_use_backup_rules,ccu_content_01,ctm_in_thread_warning_show_report,debug_logs_defaulted_on_android,dialtone_android_eligibility,disable_zero_h_conditional_worker,disable_zero_optin_conditional_worker,disable_zero_token_conditional_worker,enable_crashreport_gk_dump,enable_crashreport_mobileconfig_dump,enable_multi_sso_logging_killswitch,enable_rewrite_for_heroplayer_killswitch,fb4a_allow_carrier_signal_on_wifi,fb4a_enable_io_logging_across_add_dexes,fb4a_internal_relogin_gk,fb4a_report_low_memory_event,fb4a_sample_prefetch_abr_at_qpl_logger,fb_app_zero_rating,fbandroid_disable_memory_trimmable,fbandroid_network_data_logger_add_enabled_features,gk_player_service_blacklist,is_employee,is_employee_public,killswitch_zero_h_ping_conditional_worker,litho_error_boundaries,ls_enable_community_messaging,m2a_overall_calling_killswitch,m4a_nux_pna_target_country,marauder_mobile_power_metrics_logging,mature_content_rating,message_attachment_size_control,messages_android_quickcam_profile,messenger_chat_head_notif_info_action_disabled,messenger_chat_heads_android,messenger_client_analytics_android,messenger_force_full_reliability_logging_android,messenger_inbox_unit_visibility_history_logging,messenger_internal_prefs_android,messenger_list_creator_debugger,messenger_marketplace_rating_use_webview,messenger_new_friend_bump_updated_launch,messenger_new_message_anchor,messenger_notification_working_group,messenger_opened_thread_from_notif_kill_switch,messenger_preload_startup_classes_asap,messenger_sms_takeover_rollout,messenger_sticker_search_android,messenger_wear_enable,mn_cowatching_android_device_eligibility,mobile_native_soft_errors,mobile_zero_show_use_data_or_stay_free_screen,mqtt_client_network_tracing,mqtt_whistle_android_unified_client_logging,msgr_android_employee_viewability_logging_debug,msgr_nux_pna_allowlist,multidexclassloader_artnative_modelspecific,orca_invite_banner_killswitch_gk,p2p_allow_product_override,p2p_android_request_eligible,p2p_android_send,p2p_android_settings,p2p_enabled_in_groups_android,p2p_has_user_added_credential_before,p2p_v2_group_requests_android,p2p_v2_local_bubble,pages_call_deflection_upsell_card_killswitch,payments_settings_p2p_entry_point,prefetch_inbox2_only_if_necessary,prefetch_thread_list_only_if_necessary,preinflate_message_item_view,rtc_android_openh264_kill_switch,rtc_coex_script_logging,rtc_fb4a_openh264_voltron_kill_switch,rtc_h264_android_device_blacklist,rtc_h265_android_device_blacklist,rtc_use_sdp_renegotiation,services_admin_export_to_calendar,skip_mount_setvisibilityhint,sms_takeover_legacy_fallback_devices,top_level_voip_call_button,use_bootstrap_zero_native,video_inline_android_shutoff,video_prefetch_fb4a,voip_audio_mode_in_call_android,voip_audio_mode_normal_android,webrtc_disable_diagnostics_folder,whistle_android,zero_backup_rewrite_rules,zero_fb4a_timespent_fix,zero_header_send_state,zero_rating_enabled_on_wifi,zero_token_header_response,zero_token_new_unknown_state_flow,zero_torque_traffic_enforcement",
                    "query_hash": "B2B524B9816B08FC2F38E8BAA941A8F9E02CF453",
                }),
                "name": "gk",
                "omit_response_on_success": False,
                "relative_url": "mobile_gatekeepers",
            },
        ]



# mautrix-instagram - A Matrix-Instagram puppeting bridge.
# Copyright (C) 2020 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from asyncpg import Connection

from mautrix.util.async_db import UpgradeTable

upgrade_table = UpgradeTable()


@upgrade_table.register(description="Latest revision", upgrades_to=8)
async def upgrade_latest(conn: Connection) -> None:
    await conn.execute(
        """CREATE TABLE portal (
            thread_id     TEXT,
            receiver      BIGINT,
            other_user_pk BIGINT,
            mxid          TEXT,
            name          TEXT,
            avatar_url    TEXT,
            name_set      BOOLEAN NOT NULL DEFAULT false,
            avatar_set    BOOLEAN NOT NULL DEFAULT false,
            encrypted     BOOLEAN NOT NULL DEFAULT false,
            relay_user_id TEXT,
            PRIMARY KEY (thread_id, receiver)
        )"""
    )
    await conn.execute(
        """CREATE TABLE "user" (
            mxid           TEXT PRIMARY KEY,
            igpk           BIGINT,
            state          jsonb,
            seq_id         BIGINT,
            snapshot_at_ms BIGINT,
            notice_room    TEXT
        )"""
    )
    await conn.execute(
        """CREATE TABLE puppet (
            pk            BIGINT PRIMARY KEY,
            name          TEXT,
            username      TEXT,
            photo_id      TEXT,
            photo_mxc     TEXT,
            name_set      BOOLEAN NOT NULL DEFAULT false,
            avatar_set    BOOLEAN NOT NULL DEFAULT false,
            is_registered BOOLEAN NOT NULL DEFAULT false,
            custom_mxid   TEXT,
            access_token  TEXT,
            next_batch    TEXT,
            base_url      TEXT
        )"""
    )
    await conn.execute(
        """CREATE TABLE user_portal (
            "user"          BIGINT,
            portal          TEXT,
            portal_receiver BIGINT,
            in_community    BOOLEAN NOT NULL DEFAULT false,
            FOREIGN KEY (portal, portal_receiver) REFERENCES portal(thread_id, receiver)
                ON UPDATE CASCADE ON DELETE CASCADE
        )"""
    )
    await conn.execute(
        """CREATE TABLE message (
            mxid     TEXT,
            mx_room  TEXT NOT NULL,
            item_id  TEXT,
            receiver BIGINT,
            sender   BIGINT NOT NULL,

            client_context TEXT,
            ig_timestamp   BIGINT,
            PRIMARY KEY (item_id, receiver),
            UNIQUE (mxid, mx_room)
        )"""
    )
    await conn.execute(
        """CREATE TABLE reaction (
            mxid         TEXT NOT NULL,
            mx_room      TEXT NOT NULL,
            ig_item_id   TEXT,
            ig_receiver  BIGINT,
            ig_sender    BIGINT,
            reaction     TEXT NOT NULL,
            mx_timestamp BIGINT,
            PRIMARY KEY (ig_item_id, ig_receiver, ig_sender),
            FOREIGN KEY (ig_item_id, ig_receiver) REFERENCES message(item_id, receiver)
                ON DELETE CASCADE ON UPDATE CASCADE,
            UNIQUE (mxid, mx_room)
        )"""
    )


@upgrade_table.register(description="Add name_set and avatar_set to portal table")
async def upgrade_v2(conn: Connection) -> None:
    await conn.execute("ALTER TABLE portal ADD COLUMN avatar_url TEXT")
    await conn.execute("ALTER TABLE portal ADD COLUMN name_set BOOLEAN NOT NULL DEFAULT false")
    await conn.execute("ALTER TABLE portal ADD COLUMN avatar_set BOOLEAN NOT NULL DEFAULT false")
    await conn.execute("UPDATE portal SET name_set=true WHERE name<>''")


@upgrade_table.register(description="Add relay user field to portal table")
async def upgrade_v3(conn: Connection) -> None:
    await conn.execute("ALTER TABLE portal ADD COLUMN relay_user_id TEXT")


@upgrade_table.register(description="Add client context field to message table")
async def upgrade_v4(conn: Connection) -> None:
    await conn.execute("ALTER TABLE message ADD COLUMN client_context TEXT")


@upgrade_table.register(description="Add ig_timestamp field to message table")
async def upgrade_v5(conn: Connection) -> None:
    await conn.execute("ALTER TABLE message ADD COLUMN ig_timestamp BIGINT")


@upgrade_table.register(description="Allow hidden events in message table")
async def upgrade_v6(conn: Connection) -> None:
    await conn.execute("ALTER TABLE message ALTER COLUMN mxid DROP NOT NULL")


@upgrade_table.register(description="Store reaction timestamps")
async def upgrade_v7(conn: Connection) -> None:
    await conn.execute("ALTER TABLE reaction ADD COLUMN mx_timestamp BIGINT")


@upgrade_table.register(description="Store sync sequence ID in user table")
async def upgrade_v8(conn: Connection) -> None:
    await conn.execute('ALTER TABLE "user" ADD COLUMN seq_id BIGINT')
    await conn.execute('ALTER TABLE "user" ADD COLUMN snapshot_at_ms BIGINT')



from .common import MessageUnsendability, ThreadFolder
from .graphql import (
    DownloadImageFragment,
    FbIdToCursorQuery,
    FetchStickersWithPreviewsQuery,
    FileAttachmentUrlQuery,
    FileAttachmentURLResponse,
    GraphQLMutation,
    GraphQLQuery,
    ImageFragment,
    MessageList,
    MessageReactionMutation,
    MessageUndoSend,
    MessageUnsendResponse,
    MoreMessagesQuery,
    NTContext,
    ReactionAction,
    SearchEntitiesNamedQuery,
    SearchEntitiesResponse,
    StickerPreviewResponse,
    SubsequentMediaQuery,
    SubsequentMediaResponse,
    ThreadListQuery,
    ThreadListResponse,
    ThreadNameMutation,
    ThreadQuery,
    ThreadQueryResponse,
)
from .login import LoginResponse, MobileConfig, PasswordKeyResponse
from .media import UploadErrorData, UploadResponse
from .mqtt import (
    MarkReadRequest,
    MessageSyncError,
    MessageSyncPayload,
    OpenedThreadRequest,
    PHPOverride,
    RealtimeClientInfo,
    RealtimeConfig,
    RegionHintPayload,
    ResumeQueueRequest,
    SendMessageRequest,
    SendMessageResponse,
    SetTypingRequest,
    TypingNotification,
)



# mautrix-facebook - A Matrix-Facebook Messenger puppeting bridge.
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Any, Dict, List, Optional, Set
import base64
import json

from attr import dataclass
import attr

from maufbapi.thrift import RecursiveType, ThriftObject, TType, autospec, field
from mautrix.types import ExtensibleEnum, SerializableAttrs, SerializableEnum

from ..common import MessageUnsendability as Unsendability
from ..graphql import ExtensibleAttachment, MontageReplyData


@autospec
@dataclass
class ThreadKey(ThriftObject):
    other_user_id: int = field(TType.I64, default=None)
    thread_fbid: int = field(TType.I64, default=None)

    @property
    def id(self) -> Optional[int]:
        if self.other_user_id:
            return self.other_user_id
        elif self.thread_fbid:
            return self.thread_fbid
        else:
            return None


class ThreadReadStateEffect(ExtensibleEnum):
    MARK_READ = 1
    MARK_UNREAD = 2
    KEEP_AS_IS = 3


@autospec
@dataclass(kw_only=True)
class IGItemIDBlob(ThriftObject):
    first_64_bits: int = field(TType.I64)
    second_64_bits: int = field(TType.I64)

    @property
    def combine(self) -> int:
        return self.first_64_bits << 64 + self.second_64_bits


@autospec
@dataclass(kw_only=True)
class ConversationID(ThriftObject):
    conversation_fbid: int = field(TType.I64, default=None)
    canonical_participant_fbids: Set[int] = field(
        TType.SET, item_type=TType.I64, factory=lambda: {}
    )


@autospec
@dataclass(kw_only=True)
class MessageMetadata(ThriftObject):
    thread: ThreadKey
    id: str
    offline_threading_id: int = field(TType.I64, default=None)
    sender: int = field(TType.I64)  # actor_fbid
    timestamp: int = field(TType.I64)
    should_buzz_device: bool = False
    admin_text: str = field(default=None)
    tags: List[str] = field(factory=lambda: [])
    thread_read_state_effect: ThreadReadStateEffect = field(
        TType.I32, default=ThreadReadStateEffect.MARK_READ
    )
    skip_bump_thread: bool = False
    skip_snippet_update: bool = False
    message_unsendability: Unsendability = field(
        TType.BINARY, default=Unsendability.DENY_FOR_NON_SENDER
    )
    snippet: str = None
    # microseconds: int32?
    # index 15: undefined
    ig_item_id_blob: IGItemIDBlob = field(index=16, default=None)
    cid: ConversationID = None

    # data: struct(map) = field(index=1001)
    # folder_id: struct(system_folder_id: ??, user_folder_id: ??)
    # non_persistent_data: struct(map)


class ImageSource(ExtensibleEnum):
    UNKNOWN = 0
    FILE = 1
    QUICKCAM_FRONT = 2
    QUICKCAM_BACK = 3


@autospec
@dataclass(kw_only=True)
class ImageInfo(ThriftObject):
    original_width: int = field(TType.I32)
    original_height: int = field(TType.I32)
    uri_map: Dict[int, str] = field(
        TType.MAP,
        key_type=TType.I32,
        default=None,
        value_type=RecursiveType(TType.BINARY, python_type=str),
    )
    image_source: ImageSource = field(TType.I32, default=ImageSource.UNKNOWN)
    raw_image_uri: str = None
    raw_image_uri_format: str = None
    animated_uri_map: Dict[int, str] = field(
        TType.MAP,
        key_type=TType.I32,
        default=None,
        index=7,
        value_type=RecursiveType(TType.BINARY, python_type=str),
    )
    image_type: str = field(default=None)
    animated_image_type: str = field(default=None)
    render_as_sticker: bool = False
    mini_preview: bytes = None
    blurred_image_uri: str = None


class VideoSource(ExtensibleEnum):
    UNKNOWN = 0
    NON_QUICKCAM = 1
    QUICKCAM = 2
    SPEAKING_STICKER = 3
    RECORDED_STICKER = 4
    VIDEO_MAIL = 5


@autospec
@dataclass(kw_only=True)
class VideoInfo(ThriftObject):
    original_width: int = field(TType.I32)
    original_height: int = field(TType.I32)
    duration_ms: int = field(TType.I32)
    thumbnail_url: str
    download_url: str
    source: VideoSource = field(TType.I32, default=VideoSource.UNKNOWN)
    rotation: int = field(TType.I32, default=0)
    loop_count: int = field(TType.I32, default=0)


@autospec
@dataclass(kw_only=True)
class AudioInfo(ThriftObject):
    is_voicemail: bool
    call_id: str = None
    url: str
    duration_ms: int = field(TType.I32)
    sampling_frequency_hz: int = field(TType.I32)
    waveform: List[float] = field(TType.LIST, item_type=TType.FLOAT)
    # message_voice_transcription


@autospec
@dataclass(kw_only=True)
class Attachment(ThriftObject):
    media_id_str: str
    mime_type: str = field(default=None)
    file_name: str = field(default=None)
    media_id: int = field(TType.I64, default=None)  # fbid
    file_size: int = field(TType.I64, default=None)
    # attribution_info: struct
    extensible_media: str = field(default=None, index=7)
    # xma_graphql: str
    # blob_graphql: str
    image_info: ImageInfo = field(default=None, index=10)
    video_info: VideoInfo = field(default=None)
    audio_info: AudioInfo = field(default=None)
    # can contain a dash_manifest key with some XML as the value
    # or fbtype key with a number as value
    extra_metadata: Dict[str, str] = field(factory=lambda: {})
    # node_media_fbid: int64
    # raven_metadata: ???
    # client_attachment_type: ???
    # raven_poll_info: ???
    # generic_data_map: ???

    # haystack_handle: index 1000: str
    # generic_metadata: dict
    # hash: str
    # encryption_key: str
    # titan_type: int32
    # other_user_fbids: list
    # mercury_json: str
    # use_ref_counting: bool

    def parse_extensible(self) -> ExtensibleAttachment:
        if not self.extensible_media:
            raise ValueError("This attachment does not contain an extensible attachment")
        data = json.loads(self.extensible_media)
        raw_media_key = f"extensible_message_attachment:{self.media_id_str}"
        expected_key = base64.b64encode(raw_media_key.encode("utf-8")).decode("utf-8").rstrip("=")
        try:
            media_data = data[expected_key]
        except KeyError:
            media_data = list(data.values())[0]
        return ExtensibleAttachment.deserialize(media_data)


@autospec
@dataclass(kw_only=True)
class Reaction(ThriftObject):
    thread: ThreadKey
    message_id: str
    # index 3: unknown int32 (zero)
    reaction_sender_id: int = field(TType.I64, index=4)
    reaction: str = field(default=None)
    message_sender_id: int = field(TType.I64)
    # index 7: unknown number as string, similar to MessageMetadata's index 3


@dataclass
class PresenceInfo(SerializableAttrs):
    user_id: int = attr.ib(metadata={"json": "u"})
    status: int = attr.ib(metadata={"json": "p"})
    last_seen: int = attr.ib(metadata={"json": "l"}, default=0)


@dataclass
class Presence(SerializableAttrs):
    updates: List[PresenceInfo] = attr.ib(metadata={"json": "list"})
    list_type: str


class MentionType(SerializableEnum):
    PERSON = "p"
    SILENT = "s"
    THREAD = "t"


@dataclass
class Mention(SerializableAttrs):
    offset: int = attr.ib(metadata={"json": "o"})
    length: int = attr.ib(metadata={"json": "l"})
    user_id: str = attr.ib(metadata={"json": "i"})
    type: MentionType = attr.ib(metadata={"json": "t"}, default=MentionType.PERSON)


@autospec
@dataclass(kw_only=True)
class Message(ThriftObject):
    metadata: MessageMetadata
    text: str = field(default=None)
    # index 3: undefined
    sticker: int = field(TType.I64, index=4, default=None)
    attachments: List[Attachment] = field(factory=lambda: [])
    # index 6: some sort of struct:
    #    1: List[BinaryThreadKey]?
    #    2: ???
    #    3: timestamp?
    #    4: timestamp?
    extra_metadata: Dict[str, bytes] = field(index=7, factory=lambda: {})

    # iris_seq_id = field(TType.I64, index=1000)
    # generic_data_map: struct(map)
    # reply_to_message_id: str
    # message_reply:
    #   reply_to_message_id: struct(id: str)
    #   status: enum(VALID=0, DELETED, TEMPORARILY_UNAVAILABLE)
    #   reply_to_item_id:
    #     offline_threading_id: str = field(index=10)
    #     client_context: str
    #     item_id: IGItemIDBlob
    # request_context: index 1012: map<binary, binary>
    #   key apiArgs: binary containing thrift
    #     index 2: binary url, https://www.facebook.com/intern/agent/realtime_delivery/
    #     index 4: int64 (ex: 0)
    #     index 7: binary, empty?
    #     index 5: binary, some sort of uuid
    #     index 8: list<map>
    #       item 1: map<binary, binary>
    #         {"layer": "www", "push_phase": "C3", "www_rev": "1003179603",
    #          "buenopath": "XRealtimeDeliveryThriftServerController:sendRealtimeDeliveryRequest:/ls_req:TASK_LABEL=SEND_MESSAGE_V{N}"}
    #     index 9: binary (ex: www)
    #     index 10: boolean (ex: false)
    # random_nonce: int32
    # participants: list<???>
    # iris_tags: list<binary>
    # meta_tags: list<???>
    # tq_seq_id: int64 (ex: 924)

    @property
    def mentions(self) -> List[Mention]:
        return [
            Mention.deserialize(item) for item in json.loads(self.extra_metadata.get("prng", "[]"))
        ]

    @property
    def montage_reply_data(self) -> Optional[MontageReplyData]:
        data = self.extra_metadata.get("montage_reply_data")
        if not data:
            return None
        return MontageReplyData.parse_json(data)


@autospec
@dataclass(kw_only=True)
class ExtendedMessage(ThriftObject):
    reply_to_message: Message
    message: Message


@autospec
@dataclass(kw_only=True)
class UnsendMessage(ThriftObject):
    thread: ThreadKey
    message_id: str
    timestamp: int = field(TType.I64)
    user_id: int = field(TType.I64)
    # index 5: unknown int64 (ex: 0)


@autospec
@dataclass(kw_only=True)
class PowerUpMessageWrap(ThriftObject):
    message: Message


@autospec
@dataclass(kw_only=True)
class PowerUpMessage(ThriftObject):
    powerup_type: int = field(TType.I32)
    wrapped: PowerUpMessageWrap


@autospec
@dataclass
class ExtendedAddMemberParticipant(ThriftObject):
    addee_user_id: int = field(TType.I64)
    adder_user_id: int = field(TType.I64)
    # index 3: unknown int32 (ex: 0)
    timestamp: int = field(TType.I64, index=4)


@autospec
@dataclass
class ExtendedAddMember(ThriftObject):
    thread: ThreadKey
    users: List[ExtendedAddMemberParticipant]


@autospec
@dataclass
class MessageSyncClientEvent(ThriftObject):
    # 1: deltaAdminAddedToGroupThread
    # 2: deltaAdminRemovedFromGroupThread
    # 3: deltaJoinableMode
    # 4: deltaApprovalMode
    # 5: deltaApprovalQueue
    # 6: deltaRtcCallData
    # 7: deltaGroupThreadDescription
    # 8: liveLocationData
    # 9: deltaPaymentPinProtectionStatusData
    reaction: Reaction = field(index=10, default=None)
    # 11: deltaRoomDiscoverableMode
    # 12: deltaThreadSnapshot
    # 13: deltaMediaUpdatesData
    # 14: deltaMuteThreadReactions
    # 15: deltaMuteThreadMentions
    # 16: deltaOmniMDirectives
    # 17: deltaMontageParticipantsUpdate
    # 18: deltaVideoRoomMode
    # 19: deltaRoomAssociatedObjectUpdate
    # 20: deltaPlatformUpdatesData
    # 21: deltaInboxUnitUpdate
    # 22: deltaReadReceipt
    # 23: deltaMontageMessageReactions
    # 24: deltaNewMontageMessage
    # 25: deltaMontageMessageDelete
    # 26: deltaMontageThreadDelete
    # 27: deltaMontageThreadForcedFetch
    # 28: deltaMontageReadReceipt
    # 29: deltaMontageMarkRead
    # 30: deltaUpdatePrivateGroupJoinableLink
    # 31: deltaActivityTabUpdatesData
    # 32: deltaAdContext
    # 33: deltaThreadStreak
    # 34: deltaChangeViewerStatus
    # 35: deltaUpdateGroupEventRSVPStatus
    # 36: deltaPageThreadFollowUpData
    # 37: deltaMontageDirectOpen
    # 38: deltaMontageDirectExpire
    # 39: deltaThreadActivityNotification
    # 40: deltaMontageDirectKeep
    # 41: deltaMuteThreadGames
    # 42: deltaPageThreadSubtitleUpdate
    # ??? extended_add_member: ExtendedAddMember = field(index=42, default=None)
    # 43: deltaUpdateNotifiedChatsReadTimestamp
    # 44: deltaPeopleTabUpdatesData
    # 45: deltaUpdateEmojiStatus
    # 46: deltaUpdateFirstUnopenedMontageDirect
    # 47: deltaUpdateGroupsSyncStatus
    # 48: deltaUpdateGroupsSyncMetadata
    # 49: deltaGroupApprovalMuteSetting
    # 50: deltaPendingFolderCountChange
    # 51: deltaOmniMDirectivesV2
    # 52: deltaParticipantsSubscribeMetadata
    # 53: deltaAdminModelStatusUpdate
    # 54: deltaMessengerCYMKData
    extended_message: ExtendedMessage = field(index=55, default=None)  # deltaMessageReply
    # 56: deltaUpdateVideoChatLink
    # 57: deltaUpdateThreadTheme
    # 58: deltaUpdatePinnedThread
    # 59: deltaRTCSignalingMessage
    # 60: deltaPageThreadAssignedAdminUpdate
    # 61: deltaPromoteGroupThreadAdmin
    # 62: deltaMessageTranslation
    # 63: deltaMessagingFolderSettingUpdate
    # 64: deltaThreadConnectivityStatusUpdate
    # 65: deltaMessengerThreadActivityBannerUpdate
    # 66: deltaLivingRoomStatusUpdate
    unsend_message: UnsendMessage = field(index=67, default=None)  # deltaRecallMessageData
    # 68: deltaMentorshipUpdate
    # 69: deltaPageUnSubscribeStatus
    # 70: deltaRTCMultiwayMessage
    # 71: deltaMessengerRelationshipEventEligibilityUpdate
    # 72: deltaSchoolChatShouldShowInviteScreenUpdate
    # 73: deltaMessengerAdsConversionUpdate
    # 74: deltaUpdateSavedMessage
    # 75: deltaTweensAnswerWouldYouRather
    # 76: deltaMessageVoiceTranscription
    # 77: deltaPageBlurredImageStatus
    # 78: deltaHideMessageForMessengerKidsData
    # 79: deltaWorkChatSuggestions
    # 80: deltaUpdatePinnedMessage
    # 81: deltaGlobalMute
    # 82: deltaSwitchAccountBadgingUpdated
    # 83: deltaUpdateThreadDisappearingMode
    # 84: deltaPolicyViolation
    # 85: deltaMessengerBusinessSuggestedReplyUpdate
    # 86: deltaMuteCallsFromThread
    # 87: deltaNewRavenMessage
    # 88: deltaRavenAction
    # 89: deltaNewFriendBumpSeen
    # 90: deltaUpdateThreadEmoji
    # 91: deltaGlobalNotificationSettingControl
    # 92: deltaGlobalNewFriendBumpSetting
    # 93: deltaMessagingReachabilitySettingUpdate
    # 94: deltaGlobalReplyReminderSetting
    powerup_message: PowerUpMessage = field(index=95, default=None)  # 95: deltaMessagePowerUp
    # 96: deltaRtcRoomData
    # 97: deltaBiiMSavedRepliesData
    # 98: deltaGlobalMessageReminderSetting
    # 99: deltaUpdateThreadSnippet
    # 100: deltaUpdateMagicWords
    # 101: deltaSecondaryLanguageBody
    # 102: deltaSoundBite
    # 103: deltaGroupThreadNotifSettings
    # 104: deltaMessengerGroupThreadWarning
    # 105: deltaLastMissedCallData
    # 106: deltaAcceptGroupThread
    # 107: deltaBiiMPageMessageNotification
    # 109: deltaParticipantSpecialThreadRole
    # 110: deltaUpdatePinnedMessagesV2
    # 111: deltaIsAllUnreadMessageMissedCallXma
    # 112: deltaThreadCutoverData
    # 113: deltaInboxPageMessageNotification
    # 115: deltaRemoveMessage


@autospec
@dataclass
class MessageSyncClientPayload(ThriftObject):
    items: List[MessageSyncClientEvent]


@autospec
@dataclass(kw_only=True)
class MessageSyncClientWrapper(ThriftObject):
    data: bytes

    def parse(self) -> MessageSyncClientPayload:
        return MessageSyncClientPayload.from_thrift(self.data)


@autospec
@dataclass
class ReadReceipt(ThriftObject):
    thread: ThreadKey
    user_id: int = field(TType.I64)
    read_at: int = field(TType.I64)
    read_to: int = field(TType.I64)


@autospec
@dataclass
class OwnReadReceipt(ThriftObject):
    threads: List[ThreadKey]
    # index 2: ???
    read_to: int = field(TType.I64, index=3)
    read_at: int = field(TType.I64)


@autospec
@dataclass
class NameChange(ThriftObject):
    metadata: MessageMetadata
    new_name: str


@autospec
@dataclass
class AvatarChange(ThriftObject):
    metadata: MessageMetadata
    new_avatar: Attachment


class ThreadChangeAction(ExtensibleEnum):
    # action_data:
    #   'thread_icon_url': 'https://www.facebook.com/images/emoji.php/v9/t54/1/16/1f408.png'
    #   'thread_icon': '🐈'
    ICON = "change_thread_icon"

    # action_data:
    #   'should_show_icon': '1'
    #   'theme_color': 'FF5E007E'
    #   'accessibility_label': 'Grape'
    THEME = "change_thread_theme"

    # action_data:
    #   'THREAD_CATEGORY': 'GROUP'
    #   'TARGET_ID': '<user id>'
    #   'ADMIN_TYPE': '0'
    #   'ADMIN_EVENT': 'add_admin' or 'remove_admin'
    ADMINS = "change_thread_admins"

    # action_data:
    #   'APPROVAL_MODE': '1' (or '0'?)
    #   'THREAD_CATEGORY': 'GROUP'
    APPROVAL_MODE = "change_thread_approval_mode"

    # action_data:
    #   'nickname': '<per-room displayname>'
    #   'participant_id': '<user id>'
    NICKNAME = "change_thread_nickname"

    # action_data:
    #    'added_option_ids': '<option ids>' (this is a string with a list inside)
    #    'removed_option_ids': '<option ids>' (this is a string with a list inside)
    #    'question_json': JSON of the question
    POLL = "group_poll"


@autospec
@dataclass(kw_only=True)
class ThreadChange(ThriftObject):
    metadata: MessageMetadata
    action: ThreadChangeAction = field(TType.BINARY)
    action_data: Dict[str, str] = field(
        TType.MAP,
        key_type=RecursiveType(TType.BINARY, python_type=str),
        value_type=RecursiveType(TType.BINARY, python_type=str),
    )


@autospec
@dataclass(kw_only=True)
class AddMemberParticipant(ThriftObject):
    id: int = field(TType.I64)
    first_name: str
    name: str
    # index 4: unknown boolean


@autospec
@dataclass(kw_only=True)
class AddMember(ThriftObject):
    metadata: MessageMetadata
    users: List[AddMemberParticipant]


@autospec
@dataclass(kw_only=True)
class RemoveMember(ThriftObject):
    metadata: MessageMetadata
    user_id: int = field(TType.I64)


@autospec
@dataclass(kw_only=True)
class DeliveryReceipt(ThriftObject):
    thread: ThreadKey
    user_id: Optional[int] = field(TType.I64, default=None)
    # indices 3-5: ???
    message_id_list: List[str] = field(index=6)
    timestamp: int = field(TType.I64)


@autospec
@dataclass(kw_only=True)
class ForcedFetch(ThriftObject):
    thread: ThreadKey
    # index 2: ???
    # index 3: unknown bool (False)


@autospec
@dataclass(kw_only=True)
class MessageSyncEvent(ThriftObject):
    # 1: no_op: struct (no fields)
    message: Message = field(index=2, default=None)
    # 3: new_group_thread
    own_read_receipt: OwnReadReceipt = field(index=4, default=None)  # deltaMarkRead
    # 5: mark_unread
    # 6: message_delete
    # 7: thread_delete
    add_member: AddMember = field(index=8, default=None)  # deltaParticipantsAddedToGroupThread
    remove_member: RemoveMember = field(index=9, default=None)  # deltaParticipantLeftGroupThread
    name_change: NameChange = field(index=10, default=None)
    avatar_change: AvatarChange = field(index=11, default=None)
    # 12: mute_settings
    # 13: thread_action
    # 14: thread_folder
    # 15: rtc_event_log
    # 16: video_call
    thread_change: ThreadChange = field(index=17, default=None)  # deltaAdminTextMessage
    forced_fetch: ForcedFetch = field(index=18, default=None)
    read_receipt: ReadReceipt = field(index=19, default=None)  # deltaReadReceipt
    # 20: broadcast_message
    # 21: mark_folder_seen
    # 22: sent_message
    # 23: pinned_groups
    # 24: page_admin_reply
    delivery_receipt: DeliveryReceipt = field(index=25, default=None)
    # 26: p2p_payment_message
    # 27: folder_count
    # 28: pages_manager_event
    # 29: notification_settings
    # 30: replace_message
    # 31: zero_rating
    # 32: montage_message
    # 33: genie_message
    # 34: generic_map_mutation
    # 35: admin_added
    # 36: admin_removed
    # 37: rtc_call_data
    # 38: joinable_mode
    # 39: approval_mode
    # 40: approval_queue
    # 41: amend_message
    client_payload: MessageSyncClientWrapper = field(index=42, default=None)
    # 43: non_persisted_payload
    # 44: group_history
    # 45: group_subscribe_metadata_sync
    # 46: create_new_user
    # 47: recall_message
    # 48: mutate_message_tags
    # 49: set_thread_metadata
    # 50: thread_history_delete
    # 51: new_raven_message
    # 52: demote_interop_thread
    # 53: ig_mark_thread_unread
    # 54: all_participants_removed
    # 55: create_reaction
    # 56: delete_reaction
    # 57: add_poll_vote
    # 58: raven_action
    # 59: create_one_on_one_thread
    # 1001: change_mailbox_status
    # 1002: global_message_delete

    def get_parts(self) -> List[Any]:
        parts = [
            self.message,
            self.own_read_receipt,
            self.add_member,
            self.remove_member,
            self.name_change,
            self.avatar_change,
            self.thread_change,
            self.forced_fetch,
            self.read_receipt,
            self.delivery_receipt,
        ]
        if self.client_payload:
            for inner_item in self.client_payload.parse().items:
                parts += [
                    inner_item.reaction,
                    inner_item.extended_message,
                    inner_item.unsend_message,
                    inner_item.powerup_message.wrapped.message
                    if inner_item.powerup_message
                    else None,
                ]
        return [part for part in parts if part is not None]


class MessageSyncError(ExtensibleEnum):
    QUEUE_OVERFLOW = "ERROR_QUEUE_OVERFLOW"
    QUEUE_UNDERFLOW = "ERROR_QUEUE_UNDERFLOW"
    QUEUE_NOT_FOUND = "ERROR_QUEUE_NOT_FOUND"


@autospec
@dataclass(kw_only=True)
class MessageSyncPayload(ThriftObject):
    items: List[MessageSyncEvent] = field(factory=lambda: [])
    first_seq_id: int = field(TType.I64, default=None)
    last_seq_id: int = field(TType.I64, default=None)
    viewer: int = field(TType.I64, default=None)
    # indices 5-10: ???
    subscribe_ok: str = field(index=11, default=None)
    error: MessageSyncError = field(TType.BINARY, default=None)


@autospec
@dataclass(kw_only=True)
class SendMessageResponse(ThriftObject):
    offline_threading_id: int = field(TType.I64)
    success: bool
    # index 3: unknown i32 present for errors
    error_message: str = field(default=None, index=4)
    # index 5: unknown boolean present for errors


@autospec
@dataclass
class RegionHint(ThriftObject):
    code: str


@autospec
@dataclass(kw_only=True)
class RegionHintPayload(ThriftObject):
    unknown_int64: int = field(TType.I64)
    region_hint_data: bytes

    @property
    def region_hint(self) -> RegionHint:
        return RegionHint.from_thrift(self.region_hint_data)


@autospec
@dataclass
class TypingNotification(ThriftObject):
    user_id: int = field(TType.I64)
    typing_status: int = field(TType.I32)



# mautrix-instagram - A Matrix-Instagram puppeting bridge.
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar

from attr import dataclass
from yarl import URL
import asyncpg

from mautrix.types import ContentURI, SyncToken, UserID
from mautrix.util.async_db import Database

fake_db = Database.create("") if TYPE_CHECKING else None


@dataclass
class Puppet:
    db: ClassVar[Database] = fake_db

    pk: int
    name: str | None
    username: str | None
    photo_id: str | None
    photo_mxc: ContentURI | None
    name_set: bool
    avatar_set: bool

    is_registered: bool

    custom_mxid: UserID | None
    access_token: str | None
    next_batch: SyncToken | None
    base_url: URL | None

    @property
    def _values(self):
        return (
            self.pk,
            self.name,
            self.username,
            self.photo_id,
            self.photo_mxc,
            self.name_set,
            self.avatar_set,
            self.is_registered,
            self.custom_mxid,
            self.access_token,
            self.next_batch,
            str(self.base_url) if self.base_url else None,
        )

    async def insert(self) -> None:
        q = (
            "INSERT INTO puppet (pk, name, username, photo_id, photo_mxc, name_set, avatar_set,"
            "                    is_registered, custom_mxid, access_token, next_batch, base_url) "
            "VALUES ($1, $2, $3, $4, $5, $6, $7, $8, $9, $10, $11, $12)"
        )
        await self.db.execute(q, *self._values)

    async def update(self) -> None:
        q = (
            "UPDATE puppet SET name=$2, username=$3, photo_id=$4, photo_mxc=$5, name_set=$6,"
            "                  avatar_set=$7, is_registered=$8, custom_mxid=$9, access_token=$10,"
            "                  next_batch=$11, base_url=$12 "
            "WHERE pk=$1"
        )
        await self.db.execute(q, *self._values)

    @classmethod
    def _from_row(cls, row: asyncpg.Record) -> Puppet:
        data = {**row}
        base_url_str = data.pop("base_url")
        base_url = URL(base_url_str) if base_url_str is not None else None
        return cls(base_url=base_url, **data)

    @classmethod
    async def get_by_pk(cls, pk: int) -> Puppet | None:
        q = (
            "SELECT pk, name, username, photo_id, photo_mxc, name_set, avatar_set, is_registered,"
            "       custom_mxid, access_token, next_batch, base_url "
            "FROM puppet WHERE pk=$1"
        )
        row = await cls.db.fetchrow(q, pk)
        if not row:
            return None
        return cls._from_row(row)

    @classmethod
    async def get_by_custom_mxid(cls, mxid: UserID) -> Puppet | None:
        q = (
            "SELECT pk, name, username, photo_id, photo_mxc, name_set, avatar_set, is_registered,"
            "       custom_mxid, access_token, next_batch, base_url "
            "FROM puppet WHERE custom_mxid=$1"
        )
        row = await cls.db.fetchrow(q, mxid)
        if not row:
            return None
        return cls._from_row(row)

    @classmethod
    async def all_with_custom_mxid(cls) -> list[Puppet]:
        q = (
            "SELECT pk, name, username, photo_id, photo_mxc, name_set, avatar_set, is_registered,"
            "       custom_mxid, access_token, next_batch, base_url "
            "FROM puppet WHERE custom_mxid IS NOT NULL"
        )
        rows = await cls.db.fetch(q)
        return [cls._from_row(row) for row in rows]



# mautrix-instagram - A Matrix-Instagram puppeting bridge.
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Dict

from ..types import AndroidExperiment, QeSyncResponse


class AndroidExperiments:
    experiments: Dict[str, AndroidExperiment]

    def __init__(self) -> None:
        self.experiments = {}

    def update(self, updated: QeSyncResponse) -> None:
        self.experiments.update({item.name: item.parse() for item in updated.experiments})



# mautrix-instagram - A Matrix-Instagram puppeting bridge.
# Copyright (C) 2020 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

import io

from .type import TType


class ThriftReader(io.BytesIO):
    prev_field_id: int
    stack: list[int]

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.prev_field_id = 0
        self.stack = []

    def _push_stack(self) -> None:
        self.stack.append(self.prev_field_id)
        self.prev_field_id = 0

    def _pop_stack(self) -> None:
        if self.stack:
            self.prev_field_id = self.stack.pop()

    def _read_byte(self, signed: bool = False) -> int:
        return int.from_bytes(self.read(1), "big", signed=signed)

    @staticmethod
    def _from_zigzag(val: int) -> int:
        return (val >> 1) ^ -(val & 1)

    def read_small_int(self) -> int:
        return self._from_zigzag(self.read_varint())

    def read_varint(self) -> int:
        shift = 0
        result = 0
        while True:
            byte = self._read_byte()
            result |= (byte & 0x7F) << shift
            if (byte & 0x80) == 0:
                break
            shift += 7
        return result

    def read_field(self) -> TType:
        byte = self._read_byte()
        if byte == 0:
            return TType.STOP
        delta = (byte & 0xF0) >> 4
        if delta == 0:
            self.prev_field_id = self._from_zigzag(self.read_varint())
        else:
            self.prev_field_id += delta
        return TType(byte & 0x0F)



# mautrix-instagram - A Matrix-Instagram puppeting bridge.
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from mauigpapi.errors import IGNotLoggedInError
from mautrix.bridge.commands import HelpSection, command_handler

from .typehint import CommandEvent

SECTION_CONNECTION = HelpSection("Connection management", 15, "")


@command_handler(
    needs_auth=False,
    management_only=True,
    help_section=SECTION_CONNECTION,
    help_text="Mark this room as your bridge notice room",
)
async def set_notice_room(evt: CommandEvent) -> None:
    evt.sender.notice_room = evt.room_id
    await evt.sender.update()
    await evt.reply("This room has been marked as your bridge notice room")


@command_handler(
    needs_auth=False,
    management_only=True,
    help_section=SECTION_CONNECTION,
    help_text="Check if you're logged into Instagram",
)
async def ping(evt: CommandEvent) -> None:
    if not await evt.sender.is_logged_in():
        await evt.reply("You're not logged into Instagram")
        return
    try:
        user_info = await evt.sender.client.current_user()
    except IGNotLoggedInError as e:
        evt.log.exception("Got error checking current user for %s", evt.sender.mxid)
        await evt.reply("You have been logged out")
        await evt.sender.logout(error=e)
    else:
        user = user_info.user
        await evt.reply(
            f"You're logged in as {user.full_name} ([@{user.username}]"
            f"(https://instagram.com/{user.username}), user ID: {user.pk})"
        )
    if evt.sender.is_connected:
        await evt.reply("MQTT connection is active")
    elif evt.sender.mqtt and evt.sender._listen_task:
        await evt.reply("MQTT connection is reconnecting")
    else:
        await evt.reply("MQTT not connected")


@command_handler(
    needs_auth=True,
    management_only=False,
    help_section=SECTION_CONNECTION,
    help_text="Reconnect to Instagram and synchronize portals",
    aliases=["sync"],
)
async def refresh(evt: CommandEvent) -> None:
    await evt.sender.refresh()
    await evt.reply("Synchronization complete")


@command_handler(
    needs_auth=True,
    management_only=False,
    help_section=SECTION_CONNECTION,
    help_text="Connect to Instagram",
    aliases=["reconnect"],
)
async def connect(evt: CommandEvent) -> None:
    if evt.sender.is_connected:
        await evt.reply("You're already connected to Instagram.")
        return
    await evt.sender.refresh(resync=False)
    await evt.reply("Restarted connection to Instagram.")


@command_handler(
    needs_auth=True,
    management_only=False,
    help_section=SECTION_CONNECTION,
    help_text="Disconnect from Instagram",
)
async def disconnect(evt: CommandEvent) -> None:
    if not evt.sender.mqtt:
        await evt.reply("You're not connected to Instagram.")
    await evt.sender.stop_listen()
    await evt.reply("Successfully disconnected from Instagram.")



# mautrix-facebook - A Matrix-Facebook Messenger puppeting bridge.
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

from uuid import uuid4

from yarl import URL
import attr

from mautrix.types import JSON

from ..types import (
    DownloadImageFragment,
    FbIdToCursorQuery,
    FetchStickersWithPreviewsQuery,
    FileAttachmentUrlQuery,
    FileAttachmentURLResponse,
    ImageFragment,
    MessageList,
    MessageReactionMutation,
    MessageUndoSend,
    MessageUnsendResponse,
    MoreMessagesQuery,
    ReactionAction,
    SearchEntitiesNamedQuery,
    SearchEntitiesResponse,
    StickerPreviewResponse,
    SubsequentMediaQuery,
    SubsequentMediaResponse,
    ThreadListQuery,
    ThreadListResponse,
    ThreadQuery,
    ThreadQueryResponse,
)
from ..types.graphql import OwnInfo, PageInfo, Thread, ThreadMessageID
from .base import BaseAndroidAPI
from .login import LoginAPI
from .post_login import PostLoginAPI
from .upload import UploadAPI


class AndroidAPI(LoginAPI, PostLoginAPI, UploadAPI, BaseAndroidAPI):
    _file_url_cache: dict[ThreadMessageID, FileAttachmentURLResponse]

    async def fetch_thread_list(self, **kwargs) -> ThreadListResponse:
        return await self.graphql(
            ThreadListQuery(**kwargs),
            response_type=ThreadListResponse,
            path=["data", "viewer", "message_threads"],
        )

    async def fetch_thread_info(self, *thread_ids: str | int, **kwargs) -> list[Thread]:
        resp = await self.graphql(
            ThreadQuery(thread_ids=[str(i) for i in thread_ids], **kwargs),
            path=["data"],
            response_type=ThreadQueryResponse,
        )
        return resp.message_threads

    async def fetch_messages(self, thread_id: int, before_time_ms: int, **kwargs) -> MessageList:
        return await self.graphql(
            MoreMessagesQuery(
                thread_id=str(thread_id), before_time_ms=str(before_time_ms), **kwargs
            ),
            path=["data", "message_thread", "messages"],
            response_type=MessageList,
        )

    async def fetch_stickers(self, ids: list[int], **kwargs) -> StickerPreviewResponse:
        kwargs["sticker_ids"] = [str(id) for id in ids]
        return await self.graphql(
            FetchStickersWithPreviewsQuery(**kwargs),
            path=["data"],
            response_type=StickerPreviewResponse,
            b=True,
        )

    async def unsend(self, message_id: str) -> MessageUnsendResponse:
        return await self.graphql(
            MessageUndoSend(
                message_id=message_id,
                client_mutation_id=str(uuid4()),
                actor_id=str(self.state.session.uid),
            ),
            path=["data", "message_undo_send"],
            response_type=MessageUnsendResponse,
        )

    async def react(self, message_id: str, reaction: str | None) -> None:
        action = ReactionAction.ADD if reaction else ReactionAction.REMOVE
        await self.graphql(
            MessageReactionMutation(
                message_id=message_id,
                reaction=reaction,
                action=action,
                client_mutation_id=str(uuid4()),
                actor_id=str(self.state.session.uid),
            ),
            response_type=JSON,
        )

    async def fetch_image(self, media_id: int | str) -> ImageFragment:
        return await self.graphql(
            DownloadImageFragment(fbid=str(media_id)),
            path=["data", "node"],
            response_type=ImageFragment,
        )

    async def fbid_to_cursor(self, thread_id: int | str, media_id: int | str) -> PageInfo:
        return await self.graphql(
            FbIdToCursorQuery(fbid=str(media_id), thread_id=str(thread_id)),
            path=["data", "message_thread", "message_shared_media", "page_info"],
            response_type=PageInfo,
        )

    async def media_query(
        self, thread_id: int | str, cursor: str | None = None
    ) -> SubsequentMediaResponse:
        return await self.graphql(
            SubsequentMediaQuery(thread_id=str(thread_id), cursor_id=cursor),
            path=["data", "message_thread", "mediaResult"],
            response_type=SubsequentMediaResponse,
        )

    async def search(self, query: str, **kwargs) -> SearchEntitiesResponse:
        return await self.graphql(
            SearchEntitiesNamedQuery(search_query=query, **kwargs),
            path=["data", "entities_named"],
            response_type=SearchEntitiesResponse,
        )

    async def get_image_url(
        self,
        message_id: str,
        attachment_id: int | str,
        preview: bool = False,
        max_width: int = 384,
        max_height: int = 480,
    ) -> str | None:
        query = {
            "method": "POST",
            "redirect": "true",
            "access_token": self.state.session.access_token,
            "mid": f"m_{message_id}",
            "aid": str(attachment_id),
        }
        if preview:
            query["preview"] = "1"
            query["max_width"] = max_width
            query["max_height"] = max_height
        headers = {
            "referer": f"fbapp://{self.state.application.client_id}/messenger_thread_photo",
            "x-fb-friendly-name": "image",
        }
        resp = await self.get(
            (self.graph_url / "messaging_get_attachment").with_query(query),
            headers=headers,
            include_auth=False,
            allow_redirects=False,
        )
        # TODO handle errors more properly?
        try:
            return resp.headers["Location"]
        except KeyError:
            return None

    async def get_file_url(
        self, thread_id: str | int, message_id: str, attachment_id: str | int
    ) -> URL | None:
        attachment_id = str(attachment_id)
        msg_id = ThreadMessageID(thread_id=str(thread_id), message_id=message_id)
        try:
            resp = self._file_url_cache[msg_id]
        except KeyError:
            resp = await self.graphql(
                FileAttachmentUrlQuery(thread_msg_id=msg_id),
                path=["data", "message"],
                response_type=FileAttachmentURLResponse,
            )
            if len(resp.blob_attachments) > 1:
                self._file_url_cache[msg_id] = resp
        for attachment in resp.blob_attachments:
            if attachment.attachment_fbid == attachment_id:
                url = URL(resp.blob_attachments[0].url)
                if url.host == "l.facebook.com":
                    url = URL(url.query["u"])
                return url
        return None

    async def get_self(self) -> OwnInfo:
        fields = ",".join(field.name for field in attr.fields(OwnInfo))
        url = (self.graph_url / str(self.state.session.uid)).with_query({"fields": fields})
        async with self.get(url) as resp:
            json_data = await self._handle_response(resp)
        return OwnInfo.deserialize(json_data)

    async def logout(self) -> bool:
        headers = {
            **self._headers,
            "x-fb-friendly-name": "logout",
        }
        req: dict[str, str] = {
            **self._params,
            "fb_api_req_friendly_name": "logout",
            "fb_api_caller_class": "AuthOperations",
        }
        resp = await self.http.post(
            url=self.b_graph_url / "auth" / "expire_session", headers=headers, data=req
        )
        resp.raise_for_status()
        return await resp.text() == "true"

    async def cdn_rmd(self, prev_token: str = "", reason: str = "TIMER_EXPIRED") -> str:
        # reasons: TIMER_EXPIRED, APP_START, APP_RESUME
        headers = {
            **self._headers,
            "content-type": "application/x-www-form-urlencoded",
            "x-fb-friendly-name": "rmd-mapfetcher",
            "x-fb-request-analytics-tags": "rmd",
            "accept-encoding": "x-fb-dz;d=1, gzip, deflate",
        }
        query = {
            "net_iface": self.state.device.net_iface,
            "reason": reason,
        }
        if prev_token:
            query["prev_token"] = prev_token
        resp = await self.http.post(
            url=(self.graph_url / "v3.2" / "cdn_rmd").with_query(query),
            headers=headers,
        )
        await self._decompress_zstd(resp)
        self.log.trace(f"cdn_rmd response: {await resp.text()}")
        json_data = await self._handle_response(resp)
        return json_data["token"]



# mautrix-facebook - A Matrix-Facebook Messenger puppeting bridge.
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Dict, List

from attr import dataclass

from maufbapi.thrift import ThriftObject, TType, autospec, field


@autospec
@dataclass(kw_only=True)
class SendMessageRequest(ThriftObject):
    # tfbid_<groupid> for groups, plain user id for users
    chat_id: str  # to
    message: str  # body
    offline_threading_id: int = field(TType.I64)
    # coordinates: struct(accuracy: str, latitude: str, longitude: str)

    # Example values:
    #   'is_in_chatheads': 'false'
    #   'ld': '{"u":1674434.........}'
    #   'entrypoint': 'messenger_inbox:in_thread'
    #   'trigger': '2:thread_list:thread' or 'thread_view_messages_fragment_unknown'
    #   'active_now': '{"is_online":"false","last_active_seconds":"1431"}'
    #
    #   'media_camera_mode': 'VIDEO'
    #   'trigger': 'thread_view_messages_fragment_unknown'
    #   'entry_point': 'THREAD_CAMERA_COMPOSER_BUTTON'
    client_tags: Dict[str, str] = field(index=5, factory=lambda: {})
    # 369239263222822 = "like"
    sticker: str = field(default=None)  # object_attachment
    # copy_message_id: str
    # copy_attachment_id: str
    media_ids: List[str] = field(default=None, index=9)  # media_attachment_ids
    # fb_trace_meta: str
    # image_type: struct(value: int)
    sender_fbid: int = field(TType.I64, index=12)
    # broadcast_recipients: dict
    # attribution_app_id: int = field(TType.I64)
    # ios_bundle_id: str
    # android_key_hash: str
    # location_attachment: struct(coordinates: struct, isCurrentLocation: bool, placeId: int64)
    ttl: int = field(TType.I32, index=18, default=0)
    # ref_code: int = field(TType.I32)
    extra_metadata: Dict[str, str] = field(index=20, default=None)  # generic_metadata
    mark_read_watermark_timestamp: int = field(TType.I64, index=21, default=0)
    # attempt_id: str
    is_dialtone: bool = field(TType.BOOL, index=23, default=True)
    msg_attempt_id: int = field(TType.I64, index=24)
    # external_attachment_url: str
    # skip_android_hash_check: bool
    # original_copy_message_id: str
    reply_to: str = field(index=28, default=None)  # reply_to_message_id
    # log_info: dict
    # is_self_forwarded: bool
    # message_powerup_data: struct(powerUpStyle: enum(NONE=0, LOVE, GIFT_WRAP, CELEBRATION, FIRE))
    # sound_bite_id: int = field(TType.I64)
    # forward_score: int = field(TType.I32)
    # is_forwarded: bool


@autospec
@dataclass(kw_only=True)
class MarkReadRequest(ThriftObject):
    receipt_type: str = "read"
    state: bool = True
    # thread_id: str
    # action_id: int64
    # sync_seq_id: int64
    group_id: int = field(TType.I64, index=6, default=None)  # thread_fbid
    user_id: int = field(TType.I64, default=None)  # other_user_fbid
    # actor_fbid: int64
    read_to: int = field(TType.I64, index=9)  # watermark_timestamp
    # titan_originated_thread_id: str
    # should_send_read_receipt: bool
    # ad_page_message_type: str
    offline_threading_id: int = field(TType.I64, index=13)  # attempt_id


@autospec
@dataclass
class ChatIDWrapper(ThriftObject):
    chat_id: str


@autospec
@dataclass
class OpenedThreadRequest(ThriftObject):
    unknown_i64: int = field(TType.I64, default=0)
    _chat_id: bytes = field(default=None)

    @property
    def chat_id(self) -> int:
        return int(ChatIDWrapper.from_thrift(self._chat_id).chat_id)

    @chat_id.setter
    def chat_id(self, value: int) -> None:
        self._chat_id = ChatIDWrapper(str(value)).to_thrift()


@autospec
@dataclass
class SetTypingRequest(ThriftObject):
    user_id: int = field(TType.I64)  # recipient
    own_id: int = field(TType.I64)  # sender
    typing_status: int = field(TType.I32)  # state
    # attribution: struct(extension_type: str, generic_attribution_type: str, in_thread_app_id: int64, page_id: int64)
    # thread_key: struct(thread_id: str, thread_type: enum(CANONICAL=0, GROUP=1))


@autospec
@dataclass(kw_only=True)
class ResumeQueueRequest(ThriftObject):
    sync_token: str = None
    last_seq_id: int = field(TType.I64)
    max_deltas_able_to_process: int = field(TType.I32, default=None)
    delta_batch_size: int = field(TType.I32, default=None)
    encoding: str = None
    queue_type: str = None
    sync_api_version: int = field(TType.I64)
    device_id: str = None
    device_params: str = None
    queue_params: str
    entity_fbid: int = field(TType.I64, default=None)
    sync_token_long: int = field(TType.I64, index=12, default=1)
    trace_id: str = None



# mautrix-facebook - A Matrix-Facebook Messenger puppeting bridge.
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

import asyncio
import base64
import hashlib
import json
import time
import unicodedata

from aiohttp import ClientConnectionError

from ..types import UploadResponse
from .base import BaseAndroidAPI

try:
    from aiohttp_socks import ProxyConnectionError, ProxyError, ProxyTimeoutError
except ImportError:

    class ProxyError(Exception):
        pass

    ProxyConnectionError = ProxyTimeoutError = ProxyError


class UploadAPI(BaseAndroidAPI):
    async def send_media(
        self,
        data: bytes,
        file_name: str,
        mimetype: str,
        offline_threading_id: int,
        chat_id: int | None = None,
        is_group: bool | None = None,
        timestamp: int | None = None,
        reply_to: str | None = None,
        caption: str | None = None,
        duration: int | None = None,
        max_attempts: int = 5,
    ) -> UploadResponse:
        # Convert file name to ASCII with some basic accent removal
        ascii_file_name = (
            unicodedata.normalize("NFKD", file_name).encode("ascii", "ignore").decode("ascii")
        )
        headers = {
            **self._headers,
            "accept-encoding": "x-fb-dz;d=1, gzip, deflate",
            "app_id": self.state.application.client_id,
            "device_id": self.state.device.uuid,
            "attempt_id": str(offline_threading_id),
            "offset": "0",
            "x-entity-length": str(len(data)),
            "x-entity-name": ascii_file_name,
            "x-entity-type": mimetype,
            "content-type": "application/octet-stream",
            "client_tags": json.dumps(
                {
                    "trigger": "2:thread_list:thread",
                    "is_in_chatheads": "false",
                    "is_in_bubbles": "false",
                }
            ),
            "original_timestamp": str(timestamp or int(time.time() * 1000)),
            "x-msgr-region": self.state.session.region_hint,
            "x-fb-friendly-name": "post_resumable_upload_session",
        }
        if chat_id:
            headers["thread_key_type"] = "GROUP" if is_group else "ONE_TO_ONE"
            headers["send_message_by_server"] = "1"
            headers["sender_fbid"] = str(self.state.session.uid)
            headers["to"] = f"tfbid_{chat_id}" if is_group else str(chat_id)
            headers["offline_threading_id"] = str(offline_threading_id)
            headers["ttl"] = "0"
            if reply_to:
                headers["replied_to_message_id"] = reply_to
            if caption:
                headers["caption"] = base64.b64encode(caption.encode("utf-8")).decode("ascii")
        else:
            headers["send_message_by_server"] = "0"
            headers["thread_type_hint"] = "thread"
        if mimetype.startswith("image/"):
            path_type = "messenger_gif" if mimetype == "image/gif" else "messenger_image"
            headers["image_type"] = "FILE_ATTACHMENT"
            headers["media_type"] = "PHOTO"
            headers["media_send_type"] = "PICK"
        elif mimetype.startswith("video/"):
            path_type = "messenger_video"
            headers["video_type"] = "FILE_ATTACHMENT"
            headers["media_type"] = "VIDEO"
            headers["media_send_type"] = "PICK"
        elif mimetype.startswith("audio/"):
            path_type = "messenger_audio"
            headers["audio_type"] = "VOICE_MESSAGE"
            headers["media_type"] = "AUDIO"
            headers["media_send_type"] = "CAPTURE"
            headers["is_voicemail"] = "0"
            if duration:
                headers["duration"] = str(duration)
        else:
            path_type = "messenger_file"
            headers["file_type"] = "FILE_ATTACHMENT"
            headers["media_type"] = "FILE"
            headers["media_send_type"] = "PICK"

        self.log.trace("Sending upload with headers: %s", headers)
        file_id = hashlib.md5(data).hexdigest() + str(offline_threading_id)
        attempt = 0
        while True:
            try:
                resp = await self.http.post(
                    self.rupload_url / path_type / file_id, headers=headers, data=data
                )
                break
            except (
                ProxyError,
                ProxyTimeoutError,
                ProxyConnectionError,
                ClientConnectionError,
                ConnectionError,
                asyncio.TimeoutError,
            ) as e:
                wait = attempt * 2
                attempt += 1
                if attempt >= max_attempts:
                    raise
                self.log.warning(
                    f"{type(e).__name__} while trying to upload media, "
                    f"retrying in {wait} seconds: {e}"
                )
                await asyncio.sleep(wait)
        json_data = await self._handle_response(resp)
        self.log.trace("Upload response: %s %s", resp.status, json_data)
        return UploadResponse.deserialize(json_data)



from .client_info import PHPOverride, RealtimeClientInfo, RealtimeConfig, UnknownStruct
from .message import (
    AddMember,
    AddMemberParticipant,
    Attachment,
    AudioInfo,
    AvatarChange,
    ExtendedAddMember,
    ExtendedAddMemberParticipant,
    ExtendedMessage,
    ForcedFetch,
    ImageInfo,
    Mention,
    MentionType,
    Message,
    MessageMetadata,
    MessageSyncClientEvent,
    MessageSyncClientPayload,
    MessageSyncError,
    MessageSyncEvent,
    MessageSyncPayload,
    NameChange,
    OwnReadReceipt,
    Presence,
    Reaction,
    ReadReceipt,
    RegionHintPayload,
    RemoveMember,
    SendMessageResponse,
    ThreadChange,
    ThreadChangeAction,
    ThreadKey,
    TypingNotification,
    UnsendMessage,
    VideoInfo,
)
from .requests import (
    MarkReadRequest,
    OpenedThreadRequest,
    ResumeQueueRequest,
    SendMessageRequest,
    SetTypingRequest,
)
